/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2006

  This file is part of wy_array.h
*/
#include <cstring>    // for memcpy,memmove

// Element type T is for POD (memcpy works) such as arithmetic types,
// pointer types,.. for those built-in types do not have the move constructor.
//
template<typename T>
class Wy__ArrayValue {
    typedef T ElemType;
    typedef size_t SizeType;
    static const SizeType MaxSizeRawMemory=Wy_Array_MaxSizeRawMemory;
    static const SizeType MinimumCapacity=Wy_Array_MinimumCapacity;
    static const SizeType DefaultSize=Wy_Array_DefaultSize;

    ElemType* _b_bgn;           // array begin pointer
    ElemType* _b_end;           // end pointer of constructed objects
    const ElemType* _b_eob;     // array end storage ptr
    //
    static SizeType _byte_dist(const ElemType* p2, const ElemType* p1)
          { return static_cast<const char*>(p2)-static_cast<const char*>(p1); };

    // [Syn] Allocate n bytes of raw memory block
    //
    // [Ret] ElemType pointer to n bytes of raw memory block
    //       NULL= Not enough memory
    //
    inline static ElemType* _alloc_raw(size_t n) WY__TSPC() {
             return reinterpret_cast<ElemType*>(::operator new(n,std::nothrow));
           };
 
    // [Syn] Free the memory block allocated by _alloc_raw(size_t)
    //
    inline static void _free_raw(ElemType* v) WY__TSPC() {
             ::operator delete(v);
           };

    // [Syn] Allocate an array of capacity cap, and copy the front
    //       part of the array from [buf, buf+n_elem)
    //
    //  Note: assert(cap>=n_elem) to call
    //
    // [Ret] ElemType pointer to such array
    //       NULL= Not enough memory
    //
    // [Depend] T(T), ~T()
    //
    static ElemType* _dup_array(const ElemType* buf, SizeType n_elem, SizeType cap)
       {
         // allocate another array of cap capacity
         ElemType *narr=_alloc_raw(cap*sizeof(ElemType));
         if(narr==NULL) {
           return(NULL);
         }
         std::memcpy(narr,buf,n_elem*sizeof(ElemType));
         return(narr);
       };

     static void _objmove(ElemType* dest, ElemType* src, SizeType n) WY__NOTHROW__
       {
        if(dest<src) {
          for( ; n; ++src,++dest,--n) {
            *dest=*src;
          }
        } else {
          if(n==0) {
            return;
          }
          --n;
          src+=n;
          dest+=n;
          for(;;) {
            *dest=*src;
            if(n==0) {
              break;
            }
            --src;
            --dest;
            --n;
          };
        }
       };

     //ElemType* end(void);  // hide the non-const version for now

  public:
    WY_THROW_REPLY;

    typedef ElemType value_type;
    typedef SizeType size_type;

    // note: for test
    //
    #ifdef WY_DEBUG
    typedef value_type* iterator;
    typedef const value_type* const_iterator;
    #endif

    static  SizeType max_capacity(void) WY__TSPC()
       { return MaxSizeRawMemory/sizeof(ElemType); };
    static SizeType min_capacity(void) WY__TSPC()
       { return(MinimumCapacity); };

    Wy__ArrayValue() WY__TSPC(Reply)
      : _b_bgn(_alloc_raw(MinimumCapacity*sizeof(ElemType))),
        _b_end(_b_bgn),
        _b_eob(_b_bgn+MinimumCapacity)
    {
      if(_b_bgn==NULL) {
        WY_THROW(typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
      }
    };

    Wy__ArrayValue(size_t n_elem, const ElemType& elem)
      : _b_bgn(NULL), _b_end(NULL), _b_eob(NULL)
    {
      const SizeType cap( n_elem<MinimumCapacity? MinimumCapacity : n_elem );
      _b_end=_b_bgn=_alloc_raw(cap*sizeof(ElemType));
      if(_b_bgn==NULL) {
        WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+n_elem;
      for(; _b_end!=_b_eob; ++_b_end) {
        *_b_end= ElemType(elem);
      }
      _b_eob=_b_bgn+cap;
    };

    Wy__ArrayValue(const Wy__ArrayValue& src)
      : _b_bgn(NULL),  _b_end(NULL), _b_eob(NULL)
    {
      WY__PCOND_CHK(src);
      const SizeType cap( src.size()> Wy_Array<ElemType>::min_capacity()?
                          src.size(): Wy_Array<ElemType>::min_capacity() );
      _b_bgn=_dup_array(src._b_bgn,src.size(),cap);
      if(_b_bgn==NULL) {
        WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+cap;
      _b_end=_b_bgn+src.size();
    };

    Wy__ArrayValue(Wy__ArrayValue& p, Wy::ByMove_t) WY__TSPC()
      {
         // move semantics of p to this
         _b_bgn= p._b_bgn;
         _b_end= p._b_end;
         _b_eob= p._b_eob;

         #ifdef WY_DEBUG
         // cleanup semantics of p
         p._b_bgn=p._b_end=NULL; p._b_eob=NULL;
         #endif
      };

    ~Wy__ArrayValue() 
    {
      _free_raw(_b_bgn);
      _b_bgn=_b_end=NULL; _b_eob=NULL;
    };

    bool is_default(void) const WY__TSPC()
      { WY__PCOND_CHK(*this);
        return _b_end==_b_bgn; 
      };
    SizeType size(void) const WY__TSPC()
      { WY__PCOND_CHK(*this);
        return _b_end-_b_bgn; 
      };

    // ** Note **
    // These members expose the address of the internal dynamic buffer

    ElemType* begin(void) WY__TSPC()
      { WY__PCOND_CHK(*this);
        return _b_bgn; };
    const ElemType* begin(void) const WY__TSPC()
      { WY__PCOND_CHK(*this);
        return _b_bgn; };

    const ElemType* end(void) const WY__TSPC()
      { WY__PCOND_CHK(*this);
        return _b_end; };

    ElemType& front(void) WY__TSPC(Reply)
      {
        WY__PCOND_CHK(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOENT) );
        }
        return *_b_bgn; 
      };
    const ElemType& front(void) const WY__TSPC(Reply)
      {
        WY__PCOND_CHK(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOENT) );
        }
        return *_b_bgn; 
      };

    ElemType& back(void) WY__TSPC(Reply)
      {
        WY__PCOND_CHK(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOENT) );
        }
        return *(_b_end-1); 
      };
    const ElemType& back(void) const WY__TSPC(Reply)
      {
        WY__PCOND_CHK(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOENT) );
        }
        return *(_b_end-1); 
      };

    // Note: operator[] checks for valid position n
    //       App. has to make sure n<size(), this could be difficult
    //       with expr like: a[i][j][k] 
    //
    const ElemType& operator[](SizeType n) const WY__TSPC(Reply)
      {
        WY__PCOND_CHK(*this);
        if(n>=SizeType(_b_end-_b_bgn)) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
        }
        return _b_bgn[n];
      };
    ElemType& operator[](SizeType n) WY__TSPC(Reply)
      {
        WY__PCOND_CHK(*this);
        if(n>=SizeType(_b_end-_b_bgn)) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
        }
        return _b_bgn[n];
      };

    // Note: capacity is intact
    void reset(void)
      {
        WY__PCOND_CHK(*this);
        _b_end=_b_bgn;
      };

    // Note: capacity is set to needed minimum if reallocation occurred,
    //       otherwise capacity is intact
    void reset(const Wy__ArrayValue& src)
      {
        WY__PCOND_CHK(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(_b_bgn==src._b_bgn) {
            return;        // v.reset(v)
          }
          ::memcpy(_b_bgn,src._b_bgn,src_size*sizeof(ElemType));
          _b_end=_b_bgn+src_size;
        } else {
          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array<ElemType>::min_capacity()?
                              src_size : Wy_Array<ElemType>::min_capacity() );
          ElemType *ptmp=_dup_array(src._b_bgn,src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
          }
  
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
      };

    const Wy__ArrayValue& operator =(const Wy__ArrayValue& rhs)
      {
        this->reset(rhs);
        return *this; 
      };
    void swap(Wy__ArrayValue& v2) WY__TSPC()
      {
       WY__PCOND_CHK(*this);
       // v.swap(v) ok
       {
        ElemType* tmp(_b_bgn);
        _b_bgn=v2._b_bgn;
        v2._b_bgn=tmp;
        tmp=_b_end;
        _b_end=v2._b_end;
        v2._b_end=tmp;
       }
       {
        const ElemType* tmp(_b_eob);
        _b_eob=v2._b_eob;
        v2._b_eob=tmp;
       }
       /* This definition may be better
       char buf[sizeof(Wy__ArrayValue)];   // alignment problem!
       new(buf) Wy__ArrayValue(v2,ByMove);
       new(v2) Wy__ArrayValue(*this,ByMove);
       new(this) Wy__ArrayValue(*((Wy__ArrayValue*)buf),ByMove);
       */
      };

    // note: argument elem is a real object (not reference type)
    //
    void push_back(const ElemType elem)
      {
       // elem in *this ok
       WY__PCOND_CHK(*this);

       if(_b_end<_b_eob) {
         *_b_end=elem;
         ++_b_end;
         return;
       } else {
         const SizeType OrgSize( this->size() );

         #ifdef WY_DEBUG
           if(_b_eob!=_b_end) {
             WY_TERMINATE("");
           }
         #endif
         // allocate another array twice as large
         const SizeType cap( OrgSize<<1 );
         
         if(cap>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
         }
         std::memcpy(ptmp,_b_bgn,sizeof(ElemType)*OrgSize);

         ptmp[OrgSize]=elem;

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         _free_raw(_b_bgn);
         _b_bgn=ptmp;
         _b_end=ptmp+(OrgSize+1);
         _b_eob=ptmp+cap;
         return;
       }
      };

    void pop_back(void)
      {
       WY__PCOND_CHK(*this);
       if(_b_end==_b_bgn) {
         return;
       }
       --_b_end;
      };

    void erase(SizeType idx, SizeType n)
      {
        WY__PCOND_CHK(*this);
        ElemType *p1(_b_bgn+idx);   // p1 -> start of erase
        if(p1>=_b_end) {
          if(p1==_b_end) {
            return;
          }
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
        }
        // assert( p1<_b_end );

        ElemType *p2=_b_end;
        if(SizeType(p2-p1)>=n) {
          if(n<=0) {
            return;
          }
          p2=p1+n;
        }
        const SizeType NumRem(_b_end-p2);
        std::memmove(p1,p2,NumRem*sizeof(ElemType));
        _b_end=p1+NumRem;
        return;
      };

    // note: argument elem is a real object (not reference type)
    //
    void insert(SizeType idx, const ElemType elem)
      {
       // elem in *this ok
       WY__PCOND_CHK(*this);
       const SizeType OrgSize( this->size() );
       if(idx>OrgSize) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
       }
       if(OrgSize<this->_capacity()) {
         ElemType *p1(_b_bgn+idx);   // insert point
         std::memmove(p1+1,p1,(_b_end-p1)*sizeof(ElemType));
         *p1=elem;
         ++_b_end;
         return;
       } else {
         #ifdef WY_DEBUG
            if(OrgSize!=this->_capacity()) {
              WY_TERMINATE("");
            }
         #endif
         // allocate another array twice as large
         const SizeType cap( OrgSize<<1 );
         if(cap>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
         }

         std::memcpy(ptmp,_b_bgn,idx*sizeof(ElemType));
         ptmp[idx]=elem;
         std::memcpy(ptmp+idx+1,_b_bgn+idx,(OrgSize-idx)*sizeof(ElemType));

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         ElemType *p2(_b_bgn);
         _b_bgn=ptmp;
         _b_end=ptmp+OrgSize+1;
         _b_eob=ptmp+cap;

         _free_raw(p2);
         return;
        }
      };

    void replace(SizeType sidx, SizeType slen, SizeType n_elem, const ElemType& elem)
      {
       WY__PCOND_CHK(*this);
       const SizeType OrgSize( this->size() );
       if(sidx>OrgSize) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
       }

       // Clamp slen to at most the existing number of elements [sidx,_b_end)
       SizeType rs= OrgSize-sidx;   
       if(slen>rs) {
         slen=rs;
       }

       rs=OrgSize-slen+n_elem;      // resultant size of this array
       if(rs>this->max_capacity()) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
       }
       if(rs<=this->_capacity()) {
         const ElemType InsElem(elem);  // in case elem in *this

         ElemType *des_ptr=_b_bgn+sidx;      // insert point
         ElemType *src_ptr(des_ptr+slen);
         #ifdef WY_DEBUG
           if(((des_ptr-_b_bgn)+n_elem+(_b_end-src_ptr))!=rs) {
             WY_TERMINATE("");
           }
         #endif
         _objmove(des_ptr+n_elem,src_ptr,_b_end-src_ptr);

         for(SizeType i=0 ; i<n_elem; ++des_ptr,++i) {
           *des_ptr=InsElem;
         }
         _b_end=_b_bgn+rs;
       } else {
         // allocate memory for the new array
         const SizeType cap( rs );
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
         }

         ElemType *p1(ptmp+sidx);   // insert point
         for(SizeType i=0; i<n_elem; ++p1,++i) {
           *p1=elem;
         }

         ElemType* des_ptr=ptmp;
         ElemType* p2=_b_bgn;
         p1=ptmp+sidx;
         // move all elements in _b_bgn to ptmp
         for(; des_ptr!=p1; ++p2,++des_ptr) {
           *des_ptr=*p2;
         }
         des_ptr+=n_elem;
         p2+=slen;
         for(; p2!=_b_end; ++p2,++des_ptr) {
           *des_ptr=*p2;
         }

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         p2=_b_bgn;
         _b_bgn=ptmp;
         _b_end=des_ptr;
         _b_eob=ptmp+cap;

         _free_raw(p2);
       };
      };

    void resize(SizeType nsize)
      {
       WY__PCOND_CHK(*this);
       if(nsize<this->size()) {
         ElemType *p1(_b_bgn+nsize);
         _b_end=p1;
         return;
       } else if(nsize>this->size()) {
         // reserve enough capacity
         const WyRet r( this->_reserve(nsize) );
         if(r!=OK) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(r) );
         }

         ElemType *p2(_b_bgn+nsize);
         for(; _b_end<p2; ++_b_end) {
           *_b_end=ElemType();
         }
       } else { 
       };
      };

    bool operator ==(const Wy__ArrayValue& rhs) const
      {
        WY__PCOND_CHK(*this);
        if(this->size()!=rhs.size()) {
          return(false);
        }
        for(const ElemType *p1(_b_bgn), *p2(rhs._b_bgn); p1!=_b_end; ++p1,++p2) {
          if(*p1==*p2) {
          } else {
            return(false);
          }
        }
        return(true);
      };
    bool operator !=(const Wy__ArrayValue& rhs) const 
      {
        WY__PCOND_CHK(*this);
        if(this->size()!=rhs.size()) {
          return(true);
        }
        for(const ElemType *p1(_b_bgn), *p2(rhs._b_bgn); p1!=_b_end; ++p1,++p2) {
          if(*p1!=*p2) {
            return(true);
          } else {
          }
        }
        return(false);
      };

    SizeType _capacity(void) const WY__TSPC()
      { WY__PCOND_CHK(*this);
        return _b_eob-_b_bgn; };

    WyRet _reserve(SizeType rcap) WY__TSPC()
      {
        WY__PCOND_CHK(*this);
        if(rcap<=this->_capacity()) {
          return(OK);
        }
        if(rcap>this->max_capacity()) {
          WY_RETURN( Wym_EFBIG );
        }
        const SizeType OrgSize(this->size());
        ElemType *ptmp=_dup_array(_b_bgn,OrgSize,rcap);
        if(ptmp==NULL) {
          WY_RETURN( Wym_ENOMEM );
        }

        _free_raw(_b_bgn);
        _b_bgn=ptmp;
        _b_end=ptmp+OrgSize;
        _b_eob=ptmp+rcap;

        return(OK);
      };
};

template<>
struct Wy_Array<char> : Wy__ArrayValue<char> {
  Wy_Array() : Wy__ArrayValue<char>() {};
  Wy_Array(size_t s,const char& elem) : Wy__ArrayValue<char>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<char>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<char>(p,t) {};
};

template<>
struct Wy_Array<bool> : Wy__ArrayValue<bool> {
  Wy_Array() : Wy__ArrayValue<bool>() {};
  Wy_Array(size_t s,const bool& elem) : Wy__ArrayValue<bool>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<bool>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<bool>(p,t) {};
};

template<>
struct Wy_Array<signed char> : Wy__ArrayValue<signed char> {
  Wy_Array() : Wy__ArrayValue<signed char>() {};
  Wy_Array(size_t s, const signed char& elem) : Wy__ArrayValue<signed char>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<signed char>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<signed char>(p,t) {};
};

template<>
struct Wy_Array<signed short int> : Wy__ArrayValue<signed short int> {
  Wy_Array() : Wy__ArrayValue<signed short int>() {};
  Wy_Array(size_t s,const signed short int& elem) : Wy__ArrayValue<signed short int>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<signed short int>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<signed short int>(p,t) {};
};

template<>
struct Wy_Array<signed int> : Wy__ArrayValue<signed int> {
  Wy_Array() : Wy__ArrayValue<signed int>() {};
  Wy_Array(size_t s,const signed int& elem) : Wy__ArrayValue<signed int>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<signed int>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<signed int>(p,t) {};
};

template<>
struct Wy_Array<signed long int> : Wy__ArrayValue<signed long int> {
  Wy_Array() : Wy__ArrayValue<signed long int>() {};
  Wy_Array(size_t s, const signed long int& elem) : Wy__ArrayValue<signed long int>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<signed long int>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<signed long int>(p,t) {};
};

template<>
struct Wy_Array<signed long long int> : Wy__ArrayValue<signed long long int> {
  Wy_Array() : Wy__ArrayValue<signed long long int>() {};
  Wy_Array(size_t s,const signed long long int& elem) : Wy__ArrayValue<signed long long int>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<signed long long int>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<signed long long int>(p,t) {};
};

template<>
struct Wy_Array<unsigned char> : Wy__ArrayValue<unsigned char> {
  Wy_Array() : Wy__ArrayValue<unsigned char>() {};
  Wy_Array(size_t s,const unsigned char& elem) : Wy__ArrayValue<unsigned char>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<unsigned char>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<unsigned char>(p,t) {};
};

template<>
struct Wy_Array<unsigned short int> : Wy__ArrayValue<unsigned short int> {
  Wy_Array() : Wy__ArrayValue<unsigned short int>() {};
  Wy_Array(size_t s,const unsigned short int& elem) : Wy__ArrayValue<unsigned short int>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<unsigned short int>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<unsigned short int>(p,t) {};
};

template<>
struct Wy_Array<unsigned int> : Wy__ArrayValue<unsigned int> {
  Wy_Array() : Wy__ArrayValue<unsigned int>() {};
  Wy_Array(size_t s,const unsigned int& elem) : Wy__ArrayValue<unsigned int>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<unsigned int>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<unsigned int>(p,t) {};
};

template<>
struct Wy_Array<unsigned long int> : Wy__ArrayValue<unsigned long int> {
  Wy_Array() : Wy__ArrayValue<unsigned long int>() {};
  Wy_Array(size_t s,const unsigned long int& elem) : Wy__ArrayValue<unsigned long int>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<unsigned long int>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<unsigned long int>(p,t) {};
};

template<>
struct Wy_Array<unsigned long long int> : Wy__ArrayValue<unsigned long long int> {
  Wy_Array() : Wy__ArrayValue<unsigned long long int>() {};
  Wy_Array(size_t s,const unsigned long long int& elem) : Wy__ArrayValue<unsigned long long int>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<unsigned long long int>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<unsigned long long int>(p,t) {};
};

template<>
struct Wy_Array<float> : Wy__ArrayValue<float> {
  Wy_Array() : Wy__ArrayValue<float>() {};
  Wy_Array(size_t s,const float& elem) : Wy__ArrayValue<float>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<float>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<float>(p,t) {};
};

template<>
struct Wy_Array<double> : Wy__ArrayValue<double> {
  Wy_Array() : Wy__ArrayValue<double>() {};
  Wy_Array(size_t s,const double& elem) : Wy__ArrayValue<double>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<double>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<double>(p,t) {};
};

template<>
struct Wy_Array<long double> : Wy__ArrayValue<long double> {
  Wy_Array() : Wy__ArrayValue<long double>() {};
  Wy_Array(size_t s,const long double& elem) : Wy__ArrayValue<long double>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<long double>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<long double>(p,t) {};
};

template<typename T>
struct Wy_Array<T*> : Wy__ArrayValue<T*> {
  Wy_Array() : Wy__ArrayValue<T*>() {};
  Wy_Array(size_t s,T* elem) : Wy__ArrayValue<T*>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<T*>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<T*>(p,t) {};
};
