/* Copyright is licensed by GNU LGPL.                         by I.J.Wang 2004
*/

#define WYLIB_SOURCE
#include "wy__hstr.h"
#include <cstdlib>
#include <cstring>
#include <limits>

// [Syn] Copy element pointed by src_ptr to position pointed by des_ptr
//       and increase src_ptr/des_ptr to the next position. This procedure
//       iterates n_size times.
//
// Note: Macro is intended for small-size string copy (or move)
//       For string copy operation make sure "des_ptr <= src_ptr" 
//
// [Ret] src_ptr+=n_size
//       des_ptr+=n_size
//
#define WY__FCOPY(des_ptr,src_ptr,n_size) \
   for(int i=n_size; i>0; --i) { *((des_ptr)++)=*((src_ptr)++); }

// [Syn] Copy element pointed by src_ptr to position pointed by des_ptr
//       and decrease src_ptr/des_ptr to the previous position. This procedure
//       iterates n_size times.
//
// Note: Macro is intended for small-size string copy (or move)
//       For string copy operation make sure "src_ptr <= des_ptr" and
//       src/des point to the last element of the string
//
// [Ret] src_ptr-=n_size
//       des_ptr-=n_size
//
#define WY__BCOPY(des_ptr,src_ptr,n_size) \
   for(int i=n_size; i>0; --i) { *((des_ptr)--)=*((src_ptr)--); }

#ifndef __USE_GNU
static void* memrchr(const void *s, int c, size_t n)
{
 const char Tch( static_cast<char>(c) );    // not really sure
 for(const char*p(static_cast<const char*>(s)+n-1); n>0; --p,--n) {
   if(*p==Tch) {
     return(const_cast<char*>(p));
   }
 }
 return(NULL);
};
#endif

const char Wy__HStr::class_name[]="Wy__HStr";

//#define WY__DUMPMSG
#ifdef WY__DUMPMSG
#include <iostream>
#endif
class Wy__HStr::Wy__Assert_Check {
  public:
    Wy__Assert_Check() {
         if(max_capacity()>static_cast<Wy__HStr::size_type>(
                                std::numeric_limits<ssize_t>::max()/2-1)) {
           WY_TERMINATE("");
         };
         if(((int)DefaultObjSize-(int)sizeof(value_type)-(int)sizeof(Wy__HStr))
            !=min_capacity()) {
           WY_TERMINATE("");  // DefaultObjSize too small
         }
         if(_memsize(0)!=DefaultObjSize) {
           WY_TERMINATE("");
         }
         if(_memsize(min_capacity())!=DefaultObjSize) {
           WY_TERMINATE("");
         }
         if(_capmem(DefaultObjSize)!=min_capacity()) {
           WY_TERMINATE("");
         }
         if(_capmem( _memsize(min_capacity()) )!=min_capacity()) {
           WY_TERMINATE("");
         }
         for(size_type i=0; i<9*BufGranBitsMask; i+=BufGranBitsMask) {
           if(_memsize(i)<i+sizeof(Wy__HStr)+sizeof(value_type)) {
#ifdef WY__DUMPMSG
             std::cout << i << std::endl;
#endif
             WY_TERMINATE("");
           }
           if(_memsize(i)<DefaultObjSize) {
#ifdef WY__DUMPMSG
             std::cout << i << std::endl;
#endif
             WY_TERMINATE("");
           }
         }
         for(size_type i=DefaultObjSize; i<9*BufGranBitsMask; i+=BufGranBitsMask) {
           if(_capmem(i)+sizeof(Wy__HStr)+sizeof(value_type)<i) {
#ifdef WY__DUMPMSG
             std::cout << i << std::endl;
#endif
             WY_TERMINATE("");
           }
         }

         // Note: min_capacity has to be enough for 1. "UTC" 2. arithmetic
         //       type representation to prevent Wym_EFBIG report
         //
         if(min_capacity()<64+1) {     // long long in radix 2
           WY_TERMINATE("");
         }

         if(max_capacity()<WY_WHAT_IS_MAX_SIZE) {
           WY_TERMINATE("");  // should be enough for wrd(..)
         }
    };
} static const wy__do_not_use_me__;

Wy__HStr* Wy__HStr::alloc(void) WY__TSPC(Reply)
{
 const size_type tsize=_memsize(min_capacity());
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=0;
 _data_buf(hptr)[0]=EOS;

#ifdef WY_DEBUG
 _data_buf(hptr)[hptr->_m_dcap]=-5;    // last char of _data_buf
 if(((char*)hptr)[tsize-1]!=-5) {      // last char of object
   WY_TERMINATE("");
 }
 _data_buf(hptr)[hptr->_m_dcap]=0x5a;  // last char of _data_buf
 if(((char*)hptr)[tsize-1]!=0x5a) {    // last char of object
   WY_TERMINATE("");
 }
#endif
 return(hptr);
};

Wy__HStr* Wy__HStr::alloc(const Wy__HStr& str) WY__TSPC(Reply)
{
 const size_type str_size(str.size());
 const size_type tsize=_memsize(str_size);
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=str_size;
 std::memcpy(_data_buf(hptr),_data_buf(&str),str_size);
 _data_buf(hptr)[str_size]=EOS;
 return(hptr);
};

Wy__HStr* Wy__HStr::alloc(size_type cap) WY__TSPC(Reply)
{
 if(cap>max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type tsize=_memsize(cap);
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=0;
 _data_buf(hptr)[0]=EOS;
 return(hptr);
};

Wy__HStr* Wy__HStr::alloc(size_type cnum, value_type ch) WY__TSPC(Reply)
{
 if(cnum>max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type tsize=_memsize(cnum);
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=cnum;
 std::memset(_data_buf(hptr),ch,cnum);
 _data_buf(hptr)[cnum]=EOS;
 return(hptr);
};

Wy__HStr* Wy__HStr::alloc(const WyCSeg& cs) WY__TSPC(Reply)
{
 const size_type csize(cs.size());
 if(csize>max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type tsize=_memsize(csize);
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=csize;
 std::memcpy(_data_buf(hptr),cs.begin(),csize);
 _data_buf(hptr)[csize]=EOS;
 return(hptr);
};

// note: buffer capacity is larger than enough by 1/4
//
Wy__HStr* Wy__HStr::alloc(const WyCSeg& cs1, const WyCSeg& cs2) WY__TSPC(Reply)
{
 const size_type csize1(cs1.size());
 const size_type csize2(cs2.size());
 if((csize1>Wy__HStr::max_capacity())||(csize2>Wy__HStr::max_capacity())) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type rs(csize1+csize2);
 if(rs>max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type tsize=_memsize(rs+(rs>>2));
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=rs;
 std::memcpy(_data_buf(hptr)       ,cs1.begin(),csize1);
 std::memcpy(_data_buf(hptr)+csize1,cs2.begin(),csize2);
 _data_buf(hptr)[rs]=EOS;
 return(hptr);
};

// note: buffer capacity is larger than enough by 1/4
//
Wy__HStr* Wy__HStr::alloc(const WyCSeg& cs1, const WyCSeg& cs2, const WyCSeg& cs3) WY__TSPC(Reply)
{
 const size_type csize1(cs1.size());
 const size_type csize2(cs2.size());
 const size_type csize3(cs3.size());

 if((csize1>Wy__HStr::max_capacity())||(csize2>Wy__HStr::max_capacity())
     ||(csize3>Wy__HStr::max_capacity())) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 size_type rs(csize1+csize2);
 if(rs>max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 rs+=csize3;
 if(rs>max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }

 const size_type tsize=_memsize(rs+(rs>>2));
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=rs;
 std::memcpy(_data_buf(hptr)              ,cs1.begin(),csize1);
 std::memcpy(_data_buf(hptr)+csize1       ,cs2.begin(),csize2);
 std::memcpy(_data_buf(hptr)+csize1+csize2,cs3.begin(),csize3);
 _data_buf(hptr)[rs]=EOS;
 return(hptr);
};

// note: buffer capacity is larger than enough by 1/4
//
Wy__HStr* Wy__HStr::alloc(const WyCSeg& cs, value_type ch) WY__TSPC(Reply)
{
 const size_type csize( cs.size() );
 if(csize>=max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type tsize=_memsize(csize+sizeof(ch)+(csize>>2));
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=csize+1;
 Wy::_strcpy(_data_buf(hptr),cs);
 _data_buf(hptr)[csize]=ch;
 _data_buf(hptr)[hptr->_m_dlen]=EOS;
 return(hptr);
};

// note: buffer capacity is larger than enough by 1/4
//
Wy__HStr* Wy__HStr::alloc(value_type ch,const WyCSeg& cs) WY__TSPC(Reply)
{
 const size_type csize( cs.size() );
 if(csize>=max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type tsize=_memsize(csize+sizeof(ch)+(csize>>2));
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=csize+1;
 _data_buf(hptr)[0]=ch;
 Wy::_strcpy(_data_buf(hptr)+1,cs);
 _data_buf(hptr)[hptr->_m_dlen]=EOS;
 return(hptr);
};

// note: buffer capacity is larger than enough by 1/4
//
Wy__HStr* Wy__HStr::alloc(const WyCSeg& cs, size_type cnum, value_type ch) WY__TSPC(Reply)
{
 const size_type csize(cs.size());
 if((csize>max_capacity())||(cnum>max_capacity())) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type rs=csize+cnum;
 if(rs>max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type tsize=_memsize(rs+(rs>>2));
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=rs;
 std::memcpy(_data_buf(hptr),cs.begin(),csize);
 std::memset(_data_buf(hptr)+csize,ch,cnum);
 _data_buf(hptr)[rs]=EOS;
 return(hptr);
};

// note: buffer capacity is larger than enough by 1/4
//
Wy__HStr* Wy__HStr::alloc(const WyCSeg& cs1, size_type cnum, value_type ch,const WyCSeg& cs2) WY__TSPC(Reply)
{
 const size_type csize1(cs1.size());
 const size_type csize2(cs2.size());

 if((csize1>max_capacity())||(cnum>max_capacity())||(csize2>max_capacity())) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 size_type rs=csize1+cnum;
 if(rs>max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 rs+=csize2;
 if(rs>max_capacity()) {
   WY_THROW( Reply(Wym_EFBIG) );
 }
 const size_type tsize=_memsize(rs+(rs>>2));
 Wy__HStr* hptr=reinterpret_cast<Wy__HStr*>( _i_alloc(tsize) );
 if(hptr==NULL) {
   WY_THROW( Reply(Wym_ENOMEM) );
 }

 hptr->_m_dcap=_capmem(tsize);
 hptr->_m_dlen=rs;
 std::memcpy(_data_buf(hptr),cs1.begin(),csize1);
 std::memset(_data_buf(hptr)+csize1,ch,cnum);
 std::memcpy(_data_buf(hptr)+csize1+cnum,cs2.begin(),csize2);
 _data_buf(hptr)[rs]=EOS;
 return(hptr);
};

WyRet Wy__HStr::set_string(const Wy__HStr& hstr) WY__TSPC()
{
 const size_type hsize(hstr._m_dlen);
 value_type* des_ptr(_data_buf(this));
 if(hstr.cseg().is_overlap( WyCSeg(des_ptr,hsize+sizeof(value_type)) )) {
   // note: overlap is reported after EFBIG
   WY_RETURN(Wym_ELOOP);
 }
 if(hsize<8) {
   const WyCSeg::value_type* src_ptr(_data_buf(&hstr));
   WY__FCOPY(des_ptr,src_ptr,hsize);
   _m_dlen=hsize;
   *des_ptr=EOS;
 } else {
   std::memcpy(des_ptr,_data_buf(&hstr),hsize);
   _m_dlen=hsize;
   des_ptr[_m_dlen]=EOS;
 }
 return(OK);
};

WyRet Wy__HStr::set_string(const WyCSeg& cs) WY__TSPC()
{
 const size_type csize(cs.size());
 if(csize<6) {
   value_type* des_ptr(_data_buf(this));
   if(cs.is_overlap( WyCSeg(des_ptr,csize+sizeof(value_type)) )) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   const WyCSeg::value_type* src_ptr(cs.begin());
   WY__FCOPY(des_ptr,src_ptr,csize);
   _m_dlen=csize;
   *des_ptr=EOS;
 } else {
   if(csize>_capacity()) {
     WY_RETURN( Wym_EFBIG );
   }
   value_type* des_ptr(_data_buf(this));
   if(cs.is_overlap( WyCSeg(des_ptr,csize+sizeof(value_type)) )) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   std::memcpy(des_ptr,cs.begin(),csize);
   _m_dlen=csize;
   des_ptr[_m_dlen]=EOS;
 }
 return(OK);
};

WyRet Wy__HStr::set_string(size_type cnum,value_type ch) WY__TSPC()
{
 if(cnum>_capacity()) {
   WY_RETURN( Wym_EFBIG );
 }
 value_type* des_ptr(_data_buf(this));
 std::memset(des_ptr,ch,cnum);
 _m_dlen=cnum;
 des_ptr[cnum]=EOS;
 return(OK);
};

WyRet Wy__HStr::append(const Wy__HStr& hstr) WY__TSPC()
{
 const size_type hsize(hstr._m_dlen);
 if(hsize<8) {
   const size_type rs(_m_dlen+hsize);
   if(rs>_capacity()) {
     WY_RETURN(Wym_EFBIG);  // overflow or size too long
   }
   value_type* des_ptr(_data_buf(this)+_m_dlen);
   if(hstr.cseg().is_overlap( WyCSeg(des_ptr,_m_dcap-_m_dlen+sizeof(value_type))) ) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   const WyCSeg::value_type* src_ptr(_data_buf(&hstr));
   WY__FCOPY(des_ptr,src_ptr,hsize);
   *des_ptr=EOS;
   _m_dlen=rs;
 } else {
   if(hsize>_capacity()) {
     WY_RETURN(Wym_EFBIG);
   }
   const size_type rs(_m_dlen+hsize);
   if(rs>_capacity()) {
     WY_RETURN(Wym_EFBIG);
   }
   value_type* des_ptr(_data_buf(this)+_m_dlen);
   if(hstr.cseg().is_overlap( WyCSeg(des_ptr,_m_dcap-_m_dlen+sizeof(value_type))) ) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   std::memcpy(des_ptr,_data_buf(&hstr),hsize);
   //_data_buf(this)[rs]=EOS;
   des_ptr[hsize]=EOS;
   _m_dlen=rs;
 }
 return(OK);
};

WyRet Wy__HStr::append(const WyCSeg& cs) WY__TSPC()
{
 const size_type csize(cs.size());
 if(csize<8) {
   const size_type rs(_m_dlen+csize);
   if(rs>_capacity()) {
     WY_RETURN(Wym_EFBIG);  // overflow or size too long
   }
   value_type* des_ptr(_data_buf(this)+_m_dlen);
   if(cs.is_overlap( WyCSeg(des_ptr,_m_dcap-_m_dlen+sizeof(value_type)))) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   const WyCSeg::value_type* src_ptr(cs.begin());
   WY__FCOPY(des_ptr,src_ptr,csize);
   *des_ptr=EOS;
   _m_dlen=rs;
 } else {
   if(csize>_capacity()) {
     WY_RETURN(Wym_EFBIG);
   }
   const size_type rs(_m_dlen+csize);
   if(rs>_capacity()) {
     WY_RETURN(Wym_EFBIG);
   }
   value_type* des_ptr(_data_buf(this)+_m_dlen);
   if(cs.is_overlap( WyCSeg(des_ptr,_m_dcap-_m_dlen+sizeof(value_type)))) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   std::memcpy(des_ptr,cs.begin(),csize);
   _m_dlen=rs;
   //_data_buf(this)[rs]=EOS;
   des_ptr[csize]=EOS;
 }
 return(OK);
};

WyRet Wy__HStr::append(size_type cnum,value_type ch) WY__TSPC()
{
 if(cnum>_capacity()) {
   WY_RETURN(Wym_EFBIG);
 }
 const size_type rs=_m_dlen+cnum;
 if(rs>_capacity()) {
   WY_RETURN(Wym_EFBIG);  // overflow or size too long
 }
 value_type* des_ptr(_data_buf(this)+_m_dlen);
 std::memset(des_ptr,ch,cnum);
 _m_dlen=rs;
 des_ptr[cnum]=EOS;
 return(OK);
};

WyRet Wy__HStr::append(value_type ch) WY__TSPC()
{
 if(_m_dlen>=_capacity()) {
   WY_RETURN( Wym_EFBIG );
 }
 value_type* const cp(_data_buf(this)+(_m_dlen++));
 cp[0]=ch;
 cp[1]=EOS;
 return(OK);
};

bool Wy__HStr::find(size_type* res,value_type ch) const WY__TSPC()
{
 const value_type* const p=static_cast<const value_type*>(std::memchr(_data_buf(this),ch,_m_dlen));
 if(p==NULL) {
   return(false);
 }
#ifdef WY_DEBUG
   if(*p!=ch) {
     WY_TERMINATE("");
   }
#endif
 if(res!=NULL) {
   *res=p-_data_buf(this);
 }
 return(true);
};

bool Wy__HStr::rfind(size_type* res,value_type ch) const WY__TSPC()
{
 const value_type* const p=static_cast<const value_type*>( memrchr(_data_buf(this),ch,_m_dlen));
 if(p==NULL) {
   return(false);
 }
#ifdef WY_DEBUG
   if(*p!=ch) {
     WY_TERMINATE("");
   }
#endif
 if(res!=NULL) {
   *res=p-_data_buf(this);
 }
 return(true);
};

bool Wy__HStr::find(size_type* res,size_type idx,value_type ch) const WY__TSPC()
{
 if(idx>=_m_dlen) {
   return(false);
 }
 const value_type* const p=static_cast<const value_type*>( std::memchr(_data_buf(this)+idx,ch,_m_dlen-idx));
 if(p==NULL) {
   return(false);
 }
#ifdef WY_DEBUG
   if(*p!=ch) {
     WY_TERMINATE("");
   }
#endif
 if(res!=NULL) {
   *res=p-_data_buf(this);
 }
 return(true);
};

bool Wy__HStr::rfind(size_type* res,size_type idx,value_type ch) const WY__TSPC()
{
 if(_m_dlen==0) {
   return(false);
 }
 const size_type rs( (idx>=_m_dlen)? _m_dlen:idx+1 );
 const value_type* const p=static_cast<const value_type*>( memrchr(_data_buf(this),ch,rs));
 if(p==NULL) {
   return(false);
 }
#ifdef WY_DEBUG
   if(*p!=ch) {
     WY_TERMINATE("");
   }
#endif
 if(res!=NULL) {
   *res=p-_data_buf(this);
 }
 return(true);
};

bool Wy__HStr::find(size_type* res,size_type idx,const WyCSeg& cs) const WY__TSPC()
{
 if(idx>=_m_dlen) {
   return(false);
 }
 const size_type csize( cs.size() );
 if(csize<=0) {
   return(false); // note: memcmpy indicates "" is found
 }

 signed int fcnt(  (signed int)_m_dlen -(signed int)idx
                  -(signed int)csize   +(signed int)1 );
 const char* const cptr( cs.begin() );
 for(size_type fidx(idx); fcnt>0; --fcnt,++fidx) {
   if(::memcmp(_data_buf(this)+fidx,cptr,csize)==0) {
     if(res!=NULL) {
       *res=fidx;
     }
     return(true);
   }
 }
 return(false);
};

bool Wy__HStr::rfind(size_type* res, const WyCSeg& cs) const WY__TSPC()
{
 const size_type csize( cs.size() );
 if(csize<=0) {
   return(false); // note: memcmpy indicates "" is found
 }
 signed int fidx= (signed int)_m_dlen -(signed int)csize;
 // note: fidx may be negative

 const char* const cptr( cs.begin() );
 for( ; fidx>=0; --fidx) {
   if(::memcmp(_data_buf(this)+fidx,cptr,csize)==0) {
     if(res!=NULL) {
       *res=fidx;
     }
     return(true);
   }
 }
 return(false);
};

bool Wy__HStr::rfind(size_type* res,size_type idx,const WyCSeg& cs) const WY__TSPC()
{
 const size_type csize( cs.size() );
 if(csize<=0) {
   return(false); // note: memcmpy indicates "" is found
 }
 signed int fidx= (signed int)_m_dlen -(signed int)csize;
 if(fidx>(signed int)idx) {
   fidx=idx;
 }
 // note: fidx may be negative

 const char* const cptr( cs.begin() );
 for( ; fidx>=0; --fidx) {
   if(::memcmp(_data_buf(this)+fidx,cptr,csize)==0) {
     if(res!=NULL) {
       *res=fidx;
     }
     return(true);
   }
 }
 return(false);
};

bool Wy__HStr::find_not(size_type* res,size_type idx,value_type ch) const WY__TSPC()
{
 for( ; idx<_m_dlen; ++idx) {
   if(_data_buf(this)[idx]!=ch) {
     if(res!=NULL) {
       *res=idx;
     }
     return(true);
   }
 }
 return(false);
};

bool Wy__HStr::rfind_not(size_type* res,size_type idx,value_type ch) const WY__TSPC()
{
 if(idx>=_m_dlen) {
   if(_m_dlen<=0) {
     return(false);
   }
   idx=_m_dlen-1;
 }
 for(signed int i=idx; i>=0; --i) {
   if(_data_buf(this)[i]!=ch) {
     if(res!=NULL) {
       *res=i;
     }
     return(true);
   }
 }
 return(false);
};

WyRet Wy__HStr::insert(size_type idx,const WyCSeg& cs) WY__TSPC()
{
 if(idx>_m_dlen) {
   WY_RETURN(Wym_EINVAL);
 }
 const size_type csize( cs.size() );
 const size_type rs=_m_dlen+csize;
 if(csize<6) {
   if(rs>_capacity()) {
     WY_RETURN(Wym_EFBIG);  // overflow
   }
   char* des_ptr( _data_buf(this)+idx );
   if(cs.is_overlap( WyCSeg(des_ptr,rs-idx+sizeof(value_type)) )) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   const size_type mv_len=_m_dlen-idx;
   if(mv_len<8) {
     value_type* s_ptr(_data_buf(this)+_m_dlen-1);  // last char of src
     value_type* d_ptr(s_ptr+csize);                // last char of des

     WY__BCOPY(d_ptr,s_ptr,mv_len);
     s_ptr=const_cast<value_type*>(cs.begin());
     WY__FCOPY(des_ptr,s_ptr,csize);
   } else {
     const value_type* s_ptr(cs.begin());
     value_type* d_ptr(des_ptr+csize);
     std::memmove(d_ptr,des_ptr,mv_len);
     WY__FCOPY(des_ptr,s_ptr,csize);
   }
 } else {
   if((csize>_capacity())||(rs>_capacity())) {
     WY_RETURN(Wym_EFBIG);
   }
   char* des_ptr( _data_buf(this)+idx );
   if(cs.is_overlap( WyCSeg(des_ptr,rs-idx+sizeof(value_type)) )) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   const size_type mv_len=_m_dlen-idx;
   if(mv_len<8) {
     value_type* s_ptr(_data_buf(this)+_m_dlen-1);  // last char of src
     value_type* d_ptr(s_ptr+csize);                // last char of des
     WY__BCOPY(d_ptr,s_ptr,mv_len);
   } else {
     std::memmove(des_ptr+csize,des_ptr,mv_len);
   }

   std::memcpy(des_ptr,cs.begin(),csize);
 }
 _m_dlen=rs;
 _data_buf(this)[rs]=EOS;
 return(OK);
};

WyRet Wy__HStr::insert(size_type idx,const Wy__HStr& str) WY__TSPC()
{
 if(idx>_m_dlen) {
   WY_RETURN(Wym_EINVAL);
 }
 const size_type csize( str.size() );
 const size_type rs=_m_dlen+csize;
 if(csize<6) {
   if(rs>_capacity()) {
     WY_RETURN(Wym_EFBIG);  // overflow
   }
   char* des_ptr( _data_buf(this)+idx );
   if(str.cseg().is_overlap( WyCSeg(des_ptr,rs-idx+sizeof(value_type)) )) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   const size_type mv_len=_m_dlen-idx;
   if(mv_len<8) {
     value_type* s_ptr(_data_buf(this)+_m_dlen-1);  // last char of src
     value_type* d_ptr(s_ptr+csize);                // last char of des

     WY__BCOPY(d_ptr,s_ptr,mv_len);
     s_ptr=const_cast<value_type*>(_data_buf(&str));
     WY__FCOPY(des_ptr,s_ptr,csize);
   } else {
     const value_type* s_ptr(_data_buf(&str));
     value_type* d_ptr(des_ptr+csize);
     std::memmove(d_ptr,des_ptr,mv_len);
     WY__FCOPY(des_ptr,s_ptr,csize);
   }
 } else {
   if(rs>_capacity()) {
     WY_RETURN(Wym_EFBIG);
   }
   char* des_ptr( _data_buf(this)+idx );
   if(str.cseg().is_overlap( WyCSeg(des_ptr,rs-idx+sizeof(value_type)) )) {
     // note: overlap is reported after EFBIG
     WY_RETURN(Wym_ELOOP);
   }
   const size_type mv_len=_m_dlen-idx;
   if(mv_len<8) {
     value_type* s_ptr(_data_buf(this)+_m_dlen-1);  // last char of src
     value_type* d_ptr(s_ptr+csize);                // last char of des
     WY__BCOPY(d_ptr,s_ptr,mv_len);
   } else {
     std::memmove(des_ptr+csize,des_ptr,mv_len);
   }

   std::memcpy(des_ptr,_data_buf(&str),csize);
 }
 _m_dlen=rs;
 _data_buf(this)[rs]=EOS;
 return(OK);
};

WyRet Wy__HStr::insert(size_type idx,size_type cnum,value_type ch) WY__TSPC()
{
 if(idx>_m_dlen) {
   WY_RETURN(Wym_EINVAL);
 }
 if(cnum>_capacity()) {
   WY_RETURN(Wym_EFBIG);
 }
 const size_type rs=_m_dlen+cnum;
 if(rs>_capacity()) {
   WY_RETURN(Wym_EFBIG);  // overflow
 }
 char* des_ptr( _data_buf(this)+idx );
 const size_type mv_len=_m_dlen-idx;
 if(mv_len<8) {
   value_type* s_ptr(_data_buf(this)+_m_dlen-1);  // last char of src
   value_type* d_ptr(s_ptr+cnum);                 // last char of des
   WY__BCOPY(d_ptr,s_ptr,mv_len);
 } else {
   std::memmove(des_ptr+cnum,des_ptr,mv_len);
 }
 std::memset(des_ptr,ch,cnum);
 _m_dlen=rs;
 _data_buf(this)[rs]=EOS;
 return(OK);
};

WyRet Wy__HStr::erase(size_type idx) WY__NOTHROW__
{
 if(idx>_m_dlen) {
   WY_RETURN(Wym_EINVAL);
 }
 _m_dlen=idx;
 _data_buf(this)[_m_dlen]=EOS;
 return(OK);
};

WyRet Wy__HStr::erase(size_type idx,size_type len) WY__TSPC()
{
 if(idx>_m_dlen) {
   WY_RETURN(Wym_EINVAL);
 }
 const size_type zidx=idx+len;
 if(zidx>=_m_dlen) {
   _m_dlen=idx;
   _data_buf(this)[_m_dlen]=EOS;
   return(OK);
 }
 std::memmove(_data_buf(this)+idx,_data_buf(this)+zidx,_m_dlen-zidx);
 _m_dlen-=len;
 _data_buf(this)[_m_dlen]=EOS;
 return(OK);
};

WyRet Wy__HStr::resize(size_type num,value_type ch) WY__TSPC()
{
 // Note: Shrink should not cause error
 if(num>_capacity()) {
   WY_RETURN(Wym_EFBIG);
 }
 if(num>_m_dlen) {
   std::memset(_data_buf(this)+_m_dlen,ch,num-_m_dlen);
 }
 _m_dlen=num;
 _data_buf(this)[_m_dlen]=EOS;
 return(OK);
};

WyRet Wy__HStr::replace(size_type idx,size_type num,const WyCSeg& cs) WY__TSPC()
{
 const size_type csize( cs.size() );
 if(csize>_capacity()) {
   WY_RETURN(Wym_EFBIG);
 }
 if(idx>_m_dlen) {
   WY_RETURN(Wym_EINVAL);
 }
 size_type idx2=idx+num;
 if(idx2>_m_dlen) {
   idx2=_m_dlen;
 }
 const size_type rsiz=((_m_dlen-idx2)+idx)+csize;
 if(rsiz>_capacity()) {
   WY_RETURN(Wym_EFBIG);
 }

 char* cptr( _data_buf(this)+idx );
 if(cs.is_overlap( WyCSeg(cptr,_m_dcap-idx+sizeof(value_type)) )) {
   // note: overlap is reported after EFBIG
   WY_RETURN(Wym_ELOOP);
 }
 std::memmove(cptr+csize,_data_buf(this)+idx2,_m_dlen-idx2);
 std::memcpy(cptr,cs.begin(),csize);
 _m_dlen=rsiz;
 _data_buf(this)[_m_dlen]=EOS;
 return(OK);
};

WyRet Wy__HStr::replace(size_type idx,size_type num,size_type cnum,value_type ch) WY__TSPC()
{
 if(cnum>_capacity()) {
   WY_RETURN(Wym_EFBIG);
 }
 if(idx>_m_dlen) {
   WY_RETURN(Wym_EINVAL);
 }
 size_type idx2=idx+num;
 if(idx2>_m_dlen) {
   idx2=_m_dlen;
 }
 const size_type rsiz=((_m_dlen-idx2)+idx)+cnum;
 if(rsiz>_capacity()) {
   WY_RETURN(Wym_EFBIG);
 }
 char* cptr( _data_buf(this)+idx );
 std::memmove(cptr+cnum,_data_buf(this)+idx2,_m_dlen-idx2);
 std::memset(cptr,ch,cnum);
 _m_dlen=rsiz;
 _data_buf(this)[_m_dlen]=EOS;
 return(OK);
};

int Wy__HStr::strcmp(const Wy__HStr& str) const WY__TSPC()
{
 const int lft=_m_dlen;
 const int rit=str._m_dlen;
 if(lft!=rit) {
   return(lft-rit);
 }
 return std::memcmp(_data_buf(this),_data_buf(&str),lft);
};

int Wy__HStr::strcmp(const WyCSeg& cs) const WY__TSPC()
{
 if(cs.size()>_capacity()) {
   return(-1);
 }
 const int lft=_m_dlen;
 const int rit=cs.size();
 if(lft!=rit) {
   return(lft-rit);
 }
 return std::memcmp(_data_buf(this),cs.begin(),lft);
};

WyRet Wy__HStr::copyto(size_type* rlen,value_type* buf,size_type blen,size_type idx) const WY__TSPC()
{
 if(buf==NULL) {
   WY_RETURN(Wym_EFAULT);
 }
 if(idx>_m_dlen) {
   WY_RETURN(Wym_EINVAL);
 }
 size_type rs=_m_dlen-idx;
 if(rs>blen) {
   rs=blen;
 }
 if( WyCSeg(buf,rs).is_overlap( WyCSeg(_data_buf(this),_m_dlen+1) ) ) {
   // note: overlap is reported after EFBIG
   WY_RETURN(Wym_ELOOP);
 }
 std::memcpy(buf,_data_buf(this)+idx,rs);
 if(rlen!=NULL) {
   *rlen=rs;
 }
 return(OK);
};

WyRet Wy__HStr::_setsize(size_t nsize) WY__NOTHROW__
{
 if(nsize>_capacity()) {
   WY_RETURN(Wym_EFBIG);
 }
 _m_dlen=nsize; 
 _data_buf(this)[_m_dlen]=EOS;
 return(OK);
};

//---------------------------------------------------------------------------
