/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2005
*/
#define WYLIB_SOURCE
#include "wy__rdque.h"
#include <limits.h>     // for PIPE_BUF
#include <string>
#include "wy_atdestroy.h"

const char Wy__RdQue::class_name[]="Wy__RdQue";

// [Syn] Initialize member _dtsize suitable for reading bf
//
// [Ret] OK
//       Wym_EBADF       Object is not valid for the operation
//       Wym_ENOMEM      Not enough memory
//
WyRet Wy__RdQue::_init_dtsize(const WyByteFlow& bf)
{
 WyFileStat stt;
 const WyRet r( bf.stat(stt) );
 if(r!=OK) {
   WY_RETURN(r);
 }
 if(stt.is_fifo()) {
   _dtsize=PIPE_BUF;
 } else {
   _dtsize=stt._blksize();
 }
 if(_dtsize<512) {
   _dtsize=512;  // in case unknown
 }
 return(OK);
};


void Wy__RdQue::_set_dtsize(Wy__RdQue& rdq, size_t size)
{
 rdq._dtsize=size;
};

Wy__RdQue::Wy__RdQue(WyByteFlow& bf)
 : _fptr(&bf), _rqbuf(), _cidx(0)
{
 WyRet r( _init_dtsize(bf) );
 if(r!=OK) {
   WY_THROW( Reply(r) );
 }
 if((r=_rqbuf._reserve(_dtsize<<1) )!=OK) {
   // _reserve for default _rdque won't have Wym_EFBIG
   WY_THROW( Reply(r) ); 
 }
};

Wy__RdQue::~Wy__RdQue() WY__TSPC()
{};

WyCSeg Wy__RdQue::qdata(void) const WY__TSPC()
{
 return _rqbuf.cseg(_cidx);
};

WyByteFlow& Wy__RdQue::rfile(void) const WY__TSPC()
{
 return *_fptr;
};

WyRet Wy__RdQue::reset(WyByteFlow& bf)
try {
 Wy_AtDestroy2<void,Wy__RdQue&,size_t> 
                               rrad(Wy__RdQue::_set_dtsize,*this,_dtsize);
 WyRet r( _init_dtsize(bf) );
 if(r!=OK) {
   WY_RETURN(r);
 }
 // for now, _rqbuf capacity is only shrinked by this member
 //
 WyStr tmp;
 if((r=tmp._reserve(_dtsize<<1) )!=OK) {
   // _reserve for default _rdque won't have Wym_EFBIG
   WY_RETURN(r);
 }
 rrad.release();
 _rqbuf.swap(tmp);
 _fptr=&bf;
 _cidx=0;
 return(OK);
}
catch(const WyStr::Reply& e) {
 WY_RETURN(e);
}
catch(const WyRet& e) {
 WY_THROW( WyRet(e) );
};

WyRet Wy__RdQue::getdata(char& chr)
try {
 if(_cidx>=_rqbuf.size()) {
   _cidx=0;
   _rqbuf.reset();

   // Read and append _dtsize data to _rqbuf
   //
   WyRet r;
   size_t n_rd;
   if((r=_fptr->read(_rqbuf,_dtsize,n_rd))!=OK) {
      WY_RETURN(r);
   }
   if(n_rd==0) {
     WY_RETURN( Wym_ENOENT );
   }
 }
 chr=_rqbuf[_cidx++];
 return(OK);
}
catch(const WyRet& e) {
 WY_THROW( WyRet(e) );
};

WyRet Wy__RdQue::getdata(WyCSeg& peek, size_t count)
try {
 WyRet r;
 size_t nlen;

 peek.reset();
 if(_cidx>=_rqbuf.size()) {
   _cidx=0;
   _rqbuf.reset();
 }
 for(;;) {
   nlen=_rqbuf.size()-_cidx;   // size of available data in the queue
   if(nlen>=count) {
     peek.reset(_rqbuf.cseg(_cidx,count));
     _cidx+=count;
     return(OK);
   }
   _rqbuf.erase(0,_cidx);
   _cidx=0;

   // Read and append _dtsize data to _rqbuf
   //
   if((r=_fptr->read(_rqbuf,_dtsize,nlen))!=OK) {
      WY_RETURN(r);
   }
   if(nlen==0) {
     peek.reset(_rqbuf.cseg());
     _cidx=_rqbuf.size();
     return(OK);
   }
 }
 // UNREACHABLE
}
catch(const WyRet& e) {
 WY_THROW( WyRet(e) );
};

WyRet Wy__RdQue::getdata(WyCSeg& peek, size_t count, char c)
try {
 WyRet r;
 size_t nlen;

 peek.reset();
 if(_cidx>=_rqbuf.size()) {
   _cidx=0;
   _rqbuf.reset();
 }
 for(size_t fidx(_cidx);;) {
   const char *fptr;
   // Find char c from fidx
   //
   nlen=_rqbuf.size()-_cidx;  // size of available data in the queue
   if(nlen>=count) {
     fptr=static_cast<const char*>(std::memchr(&_rqbuf[fidx],c,count));
     if(fptr==NULL) {
       peek.reset(_rqbuf.cseg(_cidx,count));
       _cidx+=count;
     } else {
       nlen=(fptr-&_rqbuf[_cidx])+1;
       peek.reset(_rqbuf.cseg(_cidx,nlen));
       _cidx+=nlen;
     }
     return(OK);
   }
   fptr=static_cast<const char*>(std::memchr(&_rqbuf[fidx],c,nlen));
   if(fptr!=NULL) {
     nlen=(fptr-&_rqbuf[_cidx])+1;
     peek.reset(_rqbuf.cseg(_cidx,nlen));
     _cidx+=nlen;
     return(OK);
   }

   _rqbuf.erase(0,_cidx);
   _cidx=0;
   fidx=_rqbuf.size();

   // Read _dtsize data to _rqbuf
   //
   if((r=_fptr->read(_rqbuf,_dtsize,nlen))!=OK) {
      WY_RETURN(r);
   }
   if(nlen==0) {
     peek.reset(_rqbuf.cseg());
     _cidx=_rqbuf.size();
     return(OK);
   }
 }
 // UNREACHABLE
}
catch(const WyRet& e) {
 WY_THROW( WyRet(e) );
};

size_t Wy__RdQue::_rdsize(void) const WY__TSPC()
{
 return _dtsize;
};

// Note: The upper limit is detected while reading (getdata, _feedque)
//
void Wy__RdQue::_rdsize(size_t count) WY__TSPC()
{
 if(count<=0) {
   return;
 }
 _dtsize=count;
};

WyRet Wy__RdQue::_feedque(size_t& n_read)
{
 if(_cidx!=0) {
   _rqbuf.erase(0,_cidx);
   _cidx=0;
 }

 // Read and append _dtsize bytes to _rqbuf
 //
 const WyRet r( _fptr->read(_rqbuf,_dtsize,n_read) );
 if(r!=OK) {
    WY_RETURN(r);
 }
 return(OK);
};

