/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003
*/

#define WYLIB_SOURCE
#include "wy_thread.h"
#include <pthread.h>
#include <new>        // bad_alloc

#ifdef WY_DEBUG
#include "wy_uty.h"
#endif

const char Wy_Thread::class_name[]="Wy_Thread";

// Because no document found for the value of ::pthread_t for no thread.
// pick 0 for it, so far so good (the value it is initialized does not work)
//
static const ::pthread_t TID_NoThread=0;

const ::pthread_t Wy_ThreadID::Default_ThreadID=TID_NoThread;

Wy_ThreadID::Wy_ThreadID() WY__TSPC() 
 : _tid(Default_ThreadID) 
{};

Wy_ThreadID::Wy_ThreadID(const Wy_ThreadID& src) WY__TSPC() 
 : _tid(src._tid) 
{};
    
Wy_ThreadID::Wy_ThreadID(::pthread_t tid) WY__TSPC()
 : _tid(tid) 
{};

Wy_ThreadID::Wy_ThreadID(Wy_ThreadID& src, Wy::ByMove_t) WY__TSPC()
 : _tid(src._tid)
{};

::pthread_t Wy_ThreadID::_pthread_id(void) const WY__TSPC()
{
 return _tid; 
};

bool Wy_ThreadID::is_default(void) const WY__TSPC()
{
 return _tid==Default_ThreadID; 
};

void Wy_ThreadID::reset(void) WY__TSPC()
{
 _tid=Default_ThreadID; 
};

void Wy_ThreadID::reset(const Wy_ThreadID& src) WY__TSPC()
{
 _tid=src._tid; 
};

void Wy_ThreadID::reset(::pthread_t tid) WY__TSPC()
{
 _tid=tid;
};

const Wy_ThreadID& Wy_ThreadID::operator =(const Wy_ThreadID& tid) WY__TSPC()
{
 _tid=tid._tid; return(*this); 
};

bool Wy_ThreadID::operator==(Wy_ThreadID id) const WY__TSPC()
{
 return _tid==id._tid; 
};

bool Wy_ThreadID::operator!=(Wy_ThreadID id) const WY__TSPC()
{
 return _tid!=id._tid; 
};

bool Wy_ThreadID::_thread_equal(Wy_ThreadID t1, Wy_ThreadID t2) WY__TSPC()
{
 return ::pthread_equal(t1._tid,t2._tid); 
};

// value of the id for wrd(..)
unsigned int Wy_ThreadID::wy_id_value(void) const WY__NOTHROW__
{
 return _tid; 
};

//-----------------------------------------------------------------
static const size_t WY__MAX_THRDS= 
#ifdef PTHREAD_THREADS_MAX
PTHREAD_THREADS_MAX<512?  PTHREAD_THREADS_MAX:512;
#else
 512;
#endif

// [Internal] Array item of Wy__TOLTab
//            Pair of Wy_ThreadID and Wy_Thread object pointer for association
//
//  Note: The association should be 1-1, or result not predictable
//
class Wy__TOLItem {
  public:
    // [Syn] construct default object
    //
    //  tol_ptr() = 0
    //  tol_tid() = Wy_ThreadID()
    //   
    Wy__TOLItem() WY__TSPC() : _tid(),_optr(0) {};
    Wy__TOLItem(const Wy__TOLItem& src) WY__TSPC() 
       : _tid(src._tid),_optr(src._optr) {};
    Wy__TOLItem(Wy_ThreadID id, const Wy_Thread* optr) WY__TSPC() 
       : _tid(id),_optr(optr) {};
    void reset(void) WY__TSPC()
         { _tid.reset(); _optr=0; };
    void reset(const Wy__TOLItem& src) WY__TSPC() 
         { _tid.reset(src._tid); _optr=src._optr; }; // self-ops ok
    void reset(Wy_ThreadID id, const Wy_Thread* optr) WY__TSPC() 
         {  _tid.reset(id); _optr=optr; };
    bool is_default(void) const WY__TSPC()
         { return _optr==0; };
    const Wy_Thread* tol_ptr(void) const WY__TSPC()
         { return _optr; };
    Wy_ThreadID tol_tid(void) const WY__TSPC()
         { return _tid; };
    const Wy__TOLItem& operator =(const Wy__TOLItem& rhs) WY__TSPC()
         { _tid.reset(rhs._tid); _optr=rhs._optr; return(*this); };
    bool operator ==(const Wy__TOLItem& rhs) WY__TSPC()
         {
           return (_optr==rhs._optr)&&(_tid==rhs._tid);
         };
    bool operator !=(const Wy__TOLItem& rhs) WY__TSPC()
         { return !this->operator==(rhs); };
  private:
    Wy_ThreadID _tid;
    const Wy_Thread* _optr;
};

// [Internal] class of set of Wy__TOL 
//
//  Note: There is exactly one instance(wy_toltab) of this class
//
//  Note: All member access should acquire the lock first
//
class Wy__TOLTab {
  public:
    Wy__TOLTab() WY__TSPC()
       : _tabsize(0) {};

    // [Syn] Get mutex of *this
    //
    // [Ret] reply
    //
    WyMutex& mutex(void) WY__TSPC()
         { return _mtx; };

    // [Syn] Get the number of entries
    //
    // [Ret] number of Wy__TOL entries *this handles
    //
    size_t size(void) const WY__TSPC()
         { return _tabsize; };

    // [Syn] Add a thread Wy_Thread pointer and Wy_ThreadID association
    //
    //       Note: 1. If optr pointed object already in *this, no effect.
    //             2. tid must be equal to optr->tmain_id()
    //                (this member does not dereference optr)
    //
    // [Ret] OK          entry added or already in *this
    //       Wym_FAULT   optr inaccessible
    //       Wym_ENOSPC  maximal number of threads reached
    //
    WyRet add(const Wy_Thread* optr,Wy_ThreadID tid)
         {
           if(optr==0) {
             return(Wym_EINVAL);
           }
           size_t i;
           for(i=0; i<_tabsize; ++i) {
             #ifndef NDEBUG
               if(_tab[i].is_default()) {
                 WY_THROW( WyRet() );
               }
             #endif
             if(_tab[i].tol_ptr()==optr) { // base ptr is equal?
               if(Wy_ThreadID::_thread_equal(tid,_tab[i].tol_tid())==false) {
                 // optr in *this, but tid inconsistent
                 WY_THROW( WyRet() );
               }
               // already in _tab (not to add into _tab)
               return(OK);
             }
           }
           if(_tabsize>=WY__MAX_THRDS) {
             WY_RETURN(Wym_ENOSPC);
           }
           _tab[_tabsize++].reset(tid,optr);
           return(OK);
         };

    // [Syn] Remove the entry whose Wy_ThreadID is identical to tid
    //       If no entry (tid,*) in *this, no effect.
    //
    void remove(Wy_ThreadID tid) WY__TSPC()
         {
           if(_tabsize<=0) {
             return;
           }
           size_t i;
           for(i=0; (i<_tabsize)&&(_tab[i].tol_tid()!=tid); ++i) {};
           if(i>=_tabsize) {
             return;        // tid not in tab
           }
           _tabsize--;
           _tab[i].reset(_tab[_tabsize]);
         };

    // [Syn] Find entry whose Wy_ThreadID is identical to tid
    //       and return the Wy_Thread* associates with it
    //
    // [Ret] The Wy_Thread* associates with tid
    //       NULL, if no item found
    //
    const Wy_Thread* find(Wy_ThreadID tid) WY__TSPC()
         {
           size_t i;
           for(i=0; (i<_tabsize)&&(_tab[i].tol_tid()!=tid); ++i) {};
           if(i>=_tabsize) {
             return(NULL);        // tid not in tab
           }
           const Wy_Thread* retv=_tab[i].tol_ptr();
           return(retv);
         };
  private:
    Wy__TOLItem _tab[WY__MAX_THRDS];
    size_t _tabsize;
    mutable WyMutex _mtx;
} static wy_toltab;

WyStr Wy::wrd(const Wy_ThreadID& tid)
try {
 WyStr v("TID_");
 // describe Wy_ThreadID in hexidecimal number 
 WyRet r=Wy::_catstr(v,tid.wy_id_value(),16);
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(v); 
}
catch(const WyStr::Reply& e) {
 if(e==Wym_ENOMEM) {
   throw;
 }
 WY_THROW( WyRet(e) );
};

//------------------------------------------------------------------------------
bool Wy_Thread::set_cancelable(bool en)
{
 int oldstate;
 const int v(::pthread_setcancelstate( en? PTHREAD_CANCEL_ENABLE: 
                                  PTHREAD_CANCEL_DISABLE , &oldstate));
 if(v!=0) {
   // man. says only parameter error. which should not happen.
   WY_THROW(WyRet(WyReply(v)));
 }
 return (oldstate==PTHREAD_CANCEL_ENABLE);
};

Wy_Thread::CancelType Wy_Thread::set_canceltype(CancelType ct)
{
 int oldtype;
 const int v(::pthread_setcanceltype(ct,&oldtype));
 if(v!=0) {
   // man. says only parameter error. which should not happen.
   WY_THROW(WyRet(WyReply(v)));
 }
 return(static_cast<CancelType>(oldtype));
};

#ifdef WY_DEBUG
 size_t wy_joined_cnt=0;
 size_t wy_created_cnt=0;
 size_t wy_act_max=0;
#endif

// tid of main (program startup thread)
//
const Wy_ThreadID Wy_Thread::wy_main_thrd_id(::pthread_self());

::pthread_once_t Wy_Thread::_once_var=PTHREAD_ONCE_INIT;

void Wy_Thread::_once_func(void) WY__TSPC()
{
 // intended for feature test code, nothing yet.
};

// [Cancel Point]
//
//  note: thread_state() cast this member to const function
//
WyRet Wy_Thread::_join_thread(WyLock& aa)
#ifdef WY_DEBUG
try
#endif
{
 WyRet r;
 if((_thrd_id==Wy_ThreadID::Default_ThreadID)||
    (_thrd_stat==Null)) {
   WY_RETURN(Wym_ESRCH);
 }
 while(_thrd_stat==Active) {
   // cancel_point
   if((r=_thrd_stat_cond.wait(aa))!=OK) {
     WY_RETURN(r); // Wym_EINVAL
   }
 }
 if(_thrd_stat!=Stopped) {
   // assert Null state, had been reset or sth else, assume the thread
   // had been joined however (unless the machine is that fast!)
   //
   return(OK);
 }
 
 int v;
 void* thrd_ret;
 if((v=::pthread_join(_thrd_id,&thrd_ret))!=0) {
   if(v==ESRCH) {
     // Assume the thread has been joined in the unlocking interval
     //
     return(OK);
   }
   // Note: Manual listed reason(one of EINVAL): Another thread is already
   //       waiting(join) on termination of _thrd_id, which should not occur
   //       here for mutex is required to get here
   //       So, EINVAL now means 'The thread has been detatched'
   //
   WY_RETURN( WyReply(v) ); 
 }
 _thrd_id=Wy_ThreadID::Default_ThreadID;  // see note7

 #ifdef WY_DEBUG
   ++wy_joined_cnt;
   if(_thrd_stat!=Stopped) {
     WY_THROW( WyRet() );
   }
 #endif
 if(thrd_ret==PTHREAD_CANCELED) {
   _exit_obj=Wym_ETHRDCAN; WY_HERE(_exit_obj);
 }
 #ifdef WY_DEBUG
 else {
   // note6
   if(reinterpret_cast<WyRet*>(thrd_ret)!=&_exit_obj) {
     WY_THROW( WyRet() );
   }
 }
 #endif
 return(OK);
}
#ifdef WY_DEBUG
catch(...) {
 Wy::cerr << "_join_thread(..) threw" << '\n';
 throw;
}
#endif
;

// [Spc] Wy_Thread relies on pthread. Default thread attributes are:
//
//   detatchstate= PTHREAD_CREATE_JOINABLE
//   schedpolicy= SCHED_OTHER (regular, non-realtime)
//   schedparam= 0
//   inheritsched= PTHREAD_EXPLICIT_SCHED
//   scope= PTHREAD_SCOPE_SYSTEM
//
Wy_Thread::Wy_Thread()
try : _thrd_stat_cond(), _thrd_stat(Null), _thrd_id(), _exit_obj(), _in_destroy(false), _mtx()
{
 WyLock aa(_mtx);
 int v;

  if((v=::pthread_once(&_once_var, _once_func))!=0) {
     WY_THROW(WyRet(WyReply(v)));             // always succeed (man)
  }

  if((v=::pthread_attr_init(&_thrd_attr))!=0) {
    //
    // manu says only ENOMEM, linux man just says errno.
    // 
    WY_THROW( Reply(WyReply(v)) );
  }
}
catch(const Wy_Thread::Reply&) {
 throw;
}
catch(const WyLock::Reply& e) {
  WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
 WY_THROW( WyRet(e) );
};

Wy_Thread::~Wy_Thread() WY__TSPC()
try {
  //  Wy_Thread destructor is non-cancelable 
  //  Because the active thread need to access *this 
  //
  NoCancel noc;
  WyLock aa(_mtx);
  WyRet r;

  if(_in_destroy==true) {
    WY_TERMINATE("");    // destroy twice
  }
  _in_destroy=true;

  int v;

  // Cancel the possibly active thread, and join the thread
  //
  if(_thrd_stat==Active) {
    //
    // Note: Wait forever if the threead is not cancellable.
    //
    while((v=::pthread_cancel(_thrd_id))!=0) {
      // ref. note5
      if(v==EAGAIN) {
        Wy__Base::yield();  // probably ok
        continue;
      }
      WY_TERMINATE("");    // ESRCH (or others not documented)
    }
    if((r=_join_thread(aa))!=OK) {
      WY_TERMINATE("");
    }
    _thrd_stat_cond.broadcast();

  } else if(_thrd_stat==Stopped) {
    // tmain() already in Stopped state. see note8
    //
    if(_thrd_id!=Wy_ThreadID::Default_ThreadID) {
      if((r=_join_thread(aa))!=OK) {
        WY_TERMINATE("");
      }
    }
  } else {
    // Null (no thread)
  };
 
  // Cleanup members that need cleanup
  //
  if((v=::pthread_attr_destroy(&_thrd_attr))!=0) {
    // manu says only EINVAL could happen
    WY_TERMINATE("");
  }
}
catch(const WyLock::Reply&) {
 WY_TERMINATE("");
}
catch(const WyRet&) {
 WY_TERMINATE("");
}
catch(const std::exception&) {
 WY_TERMINATE("");
}
catch(...) {
 WY_TERMINATE("");
};

bool Wy_Thread::is_default(void) const
try {
 //
 // _thrd_stat==Null ensures the default state, including
 //   
 //   1. _thrd_id==Wy_ThreadID::Default_ThreadID
 //   2. _exit_obj.is_default()
 //   3. _in_destroy=false
 //   4. _thrd_attr is default
 //
 WyLock aa(_mtx);
 return(_thrd_stat==Null);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

Wy_ThreadID Wy_Thread::tmain_id(void) const
try {
 WyLock aa(_mtx);
 return Wy_ThreadID(_thrd_id); 
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// Note: This member is const, but _thrd_id may be set to default
//
Wy_Thread::ThreadState Wy_Thread::thread_state(void) const
try {
 WyLock aa(_mtx);
 if(_thrd_stat==Stopped) {
   if(_thrd_id!=Wy_ThreadID::Default_ThreadID) {
     WyRet r;
     NoCancel noc;
     if((r=const_cast<Wy_Thread*>(this)->_join_thread(aa))!=OK) {
       WY_THROW(r);
     }
   }
   return(Stopped);
 }
 return _thrd_stat;
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

WyRet Wy_Thread::exitcode(void) const
try {
 WyLock aa(_mtx);
 return(_exit_obj);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

WyRet Wy_Thread::reset(void)
try {
  WyLock aa(_mtx);

  if(_in_destroy==true) {
    WY_THROW( WyRet() );
  }

  WyRet r;
  int v;
  // If the created thread is active(running), cancel it.
  //
  if(_thrd_stat==Active) {

    if(::pthread_equal(_thrd_id,::pthread_self())) {
      // thread (tmain) is not allowed to reset itself
      // rule of libwy says reset() always returns with default object
      //
      WY_THROW( WyRet() );
    }

    NoCancel noc;
    while((v=::pthread_cancel(_thrd_id))!=0) {
      // ref. note5
      if(v==EAGAIN) {
        Wy__Base::yield();
        continue;
      }
      WY_THROW( WyRet(WyReply(v)) ); // ESRCH (or others not documented)
    }
    if((r=_join_thread(aa))!=OK) {
      WY_THROW(r);
    }

  } else if(_thrd_stat==Stopped) {
    NoCancel noc;
    if(_thrd_id!=Wy_ThreadID::Default_ThreadID) {
      if((r=_join_thread(aa))!=OK) {
        WY_THROW(r);
      }
    }
  } else {
    // Null (no thread)
  }
 
  // Reset thread attr object (destoy and initialize again)
  //
  if((v=::pthread_attr_destroy(&_thrd_attr))!=0) {
    // manu says only EINVAL could happen
    WY_THROW( WyRet(WyReply(v)) );
  }
  if((v=::pthread_attr_init(&_thrd_attr))!=0) {
    //
    // manu says only ENOMEM, linux man just says errno.
    // 
    ::pthread_attr_destroy(&_thrd_attr); // make this member invalid (release)
    WY_THROW( WyRet(WyReply(v)) );
  }

  _thrd_stat=Null;
  _thrd_id=Wy_ThreadID::Default_ThreadID;
  _exit_obj.reset();
  _thrd_stat_cond.broadcast();
  return(OK);
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const std::exception&) {
#ifdef WY_DEBUG
 Wy::cerr << "std::exception\n";
#endif
 WY_THROW( WyRet() );
};

// [Not Cancelable]
//
// [Syn] Cleanup function of the thread (return/throw)
//
//  Note: This member locks wy_toltab.mutex() and this->_mtx
//
void Wy_Thread::_tmain_close(const WyRet* r_ptr) WY__TSPC()
try {
  // thread cancelability disabled,  no need to recover, as going to end.
  //
  if(::pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL)!=0) {
    WY_TERMINATE("");  // see note2, the entire program may suffer if got this wrong.
  }
  WyLock aa_toltab( wy_toltab.mutex() );   // lock toltab first
  WyLock aa_this(_mtx);

  if(r_ptr!=0) {
    _exit_obj=*r_ptr;
  }

  // Set the thread state to Stopped and boradcast, ref. note8
  //  (destructor and reset() may wait on this)
  //
  _thrd_stat=Stopped;
  _thrd_stat_cond.broadcast();
 
  // Decrease the number of total active(running) threads
  //
  wy_toltab.remove(Wy_ThreadID(_thrd_id));
}
catch(const WyLock::Reply&) {
  WY_TERMINATE(""); // so we can trace the line
}
catch(const WyRet&) {
  WY_TERMINATE(""); // so we can trace the line
}
catch(const std::bad_alloc&) {
  WY_TERMINATE(""); // so we can trace the line
}
catch(const std::exception&) {
  WY_TERMINATE(""); // so we can trace the line
}
catch(...) {
  WY_TERMINATE(""); // so we can trace the line
};

// [Cancel Point]
//
// [Syn] The created thread function. tmain() is invoked from inside.
//
// [Ret] Pointer to _exit_obj
//
void * Wy_Thread::wy_thread_begin_func(Wy_Thread *obj_ptr)
try {

  // Syncronize with begin() by waiting for Active state
  // begin() may cancel this thread because it may fail in latter
  // processing after the thread had initiated
  //
  {
    WyLock aa(obj_ptr->_mtx);
    #ifndef NDEBUG
     if(obj_ptr->_thrd_id!=::pthread_self()) {  // assertion check
       WY_TERMINATE("");
     }
    #endif
    while(obj_ptr->_thrd_stat!=Active) {
      WyRet r;
      // This is the cancel point for begin
      //
      if((r=obj_ptr->_thrd_stat_cond.wait(aa))!=OK) {
        // Wym_EINVAL or others undocumented
        // Note: not to call _tmain_close(&r), for wy_toltab not setup in
        //       begin(). So the thread just terminates, begin() should 
        //       handle the rest
        throw(r);
      }
    }
  }

  try {
    const WyRet r( obj_ptr->tmain() );
    obj_ptr->_tmain_close(&r);
  }
  catch(...) {
    obj_ptr->_tmain_close(0);
    throw;
  };
  return(&obj_ptr->_exit_obj);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 throw;
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 throw;
}
catch(const std::bad_alloc&) {
#ifdef WY_DEBUG
 Wy::cerr << "std::bad_alloc" << '\n';
#endif
 throw;
}
catch(const std::exception&) {
#ifdef WY_DEBUG
 Wy::cerr << "std::exception" << '\n';
#endif
 throw;
}
catch(...) {
//#ifdef WY_DEBUG
// Wy::cerr << "Unknown stack unwound (maybe cancellation initiated)" << '\n';
//#endif
 throw;       // This may be cancellation initiated. If not, std::terminate()
              // may result
};

extern "C" inline void * wy__thread_begin_func_c__(void* obj_ptr)
{
  return Wy_Thread::wy_thread_begin_func(
                    reinterpret_cast<Wy_Thread *>(obj_ptr));
};

WyRet Wy_Thread::begin(void)
try {
  WyRet r;
  NoCancel noc;

  WyLock aa_toltab( wy_toltab.mutex() );   // lock toltab first
  WyLock aa_this(_mtx);
  int v;

  if(_in_destroy==true) {
    WY_THROW( WyRet() );
  }

  if(_thrd_stat!=Null) {
    WY_RETURN(Wym_EBADF);
  }
  if(wy_toltab.size()>=WY__MAX_THRDS) {
    WY_RETURN(Wym_ENFILE);
  }

  if((v=::pthread_create(&_thrd_id,&_thrd_attr, 
                         wy__thread_begin_func_c__,
                         static_cast<void*>(this)))!=0) {
     // ENOMEM can return, which is not in the manpage
     //
     WY_RETURN( WyRet(WyReply(v)) );
  }
  #ifdef WY_DEBUG
   {
     ++wy_created_cnt;
     if(wy_created_cnt<wy_joined_cnt) {
       WY_TERMINATE("");
     }
     const size_t act=wy_created_cnt-wy_joined_cnt;
     if(act>wy_act_max) {
       wy_act_max=act;
     }
   }
  #endif

  try {
    // Check that _thrd_id is not the value identical to the default id
    // Refer to definition of Wy_ThreadID::Default_ThreadID.
    //
    if(_thrd_id==Wy_ThreadID::Default_ThreadID) {
      WY_THROW( WyRet() ); // assertion failure
    }
    WyRet r( wy_toltab.add(this,_thrd_id) );
    if(r!=OK) {
      WY_THROW(r);    // no error with wy_toltab.add(..) is expected
    }
  }
  catch(...) {
    while((v=::pthread_cancel(_thrd_id))!=0) {
      // ref. note5
      if(v==EAGAIN) {
        continue;     // no yield();
      }
      WY_THROW(WyRet(WyReply(v)));    // ESRCH (or others not documented)
    }
    if((r=_join_thread(aa_this))!=OK) {
      WY_THROW(r);
    }

    _thrd_id=Wy_ThreadID::Default_ThreadID;
    _thrd_stat=Null;   // note: class data members might still inconsistent
    _exit_obj.reset(); //       but this is cahtch handler
    _thrd_stat_cond.broadcast();
    throw;
  };

  _thrd_stat=Active;
  _thrd_stat_cond.broadcast();
  return(OK);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
  WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
  WY_THROW( WyRet(e) );
};

WyRet Wy_Thread::cancel(void)
try {
  WyLock aa(_mtx);

  if(_in_destroy==true) {
    WY_THROW( WyRet() );
  }

  if(::pthread_equal(_thrd_id,::pthread_self())) {
    WY_RETURN(Wym_EPERM);  // thread can not cancel itself
  }
  if(_thrd_stat!=Active) {
    WY_RETURN(Wym_ENOENT);
  }

  int v;
  while((v=::pthread_cancel(_thrd_id))!=0) {
    // ref. note5
    if(v==EAGAIN) {
      Wy__Base::yield();
      continue;
    }
    WY_THROW( WyRet(WyReply(v)) ); // ESRCH (or others not documented)
  }
  return(OK);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// [Cancel Point]
//
WyRet Wy_Thread::wait_not(ThreadState state)
try {
 WyLock aa(_mtx);
 WyRet r;

 if(_thrd_stat==Active) {
   if(::pthread_equal(_thrd_id,::pthread_self())) {
     WY_RETURN(Wym_EPERM);  // thread cannot wait for the state of itself
   }
 }

 switch(state) {
   case Active: {
               // cancel point
               //
               // note: all returned error can be interpreted OK for this
               //       function. Along with OK, _thrd_stat cannot be Active
               //
               _join_thread(aa);

               #ifdef WY_DEBUG
                 if(_thrd_stat==Active) {
                   WY_TERMINATE("");
                 }
               #endif
              }; break;
   case Null: {
               while(_thrd_stat==Null) {
                 // cancel_point
                 if((r=_thrd_stat_cond.wait(aa))!=OK) {
                   WY_THROW( WyRet(r) ); // Wym_EINVAL
                 }
               }
              }; break;
   case Stopped: {
               if(_thrd_id!=Wy_ThreadID::Default_ThreadID) {  // see note7
                 if((r=_join_thread(aa))!=OK) {
                   WY_THROW(r);
                 }
               }
               while(_thrd_stat==Stopped) {
                 // cancel_point
                 if((r=_thrd_stat_cond.wait(aa))!=OK) {
                   WY_THROW( WyRet(r) ); // Wym_EINVAL
                 }
               }
              }; break;
   default:
        WY_THROW( WyRet() );
 };

 return(OK);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

Wy_ThreadID Wy_Thread::thread_self(void)
{
 return Wy_ThreadID(::pthread_self()); 
};

// [Cancel Point]
//
void Wy_Thread::cancel_point(void)
{
 ::pthread_testcancel(); 
};

WyRet Wy_Thread::yield(void)
{
 const int v=Wy__Base::yield();
 if(v==0) {
   return(OK);
 }
 WY_RETURN(WyReply(v));
};

// Note: speed remains an issue with this member
//
Wy_Thread* Wy_Thread::tbase_ptr(void)
try {
 WyRet r;
 if((r=wy_toltab.mutex().lock())!=OK) {
   WY_THROW(r);
 }
 const Wy_Thread* retv( wy_toltab.find(Wy_Thread::thread_self()) );
 if((r=wy_toltab.mutex().unlock())!=OK) {
   WY_THROW(r);
 }
 return(const_cast<Wy_Thread*>(retv));
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

void Wy_Thread::exit(WyRet exit_ret)
try {
 WyRet r;
 const Wy_ThreadID ctid( Wy_Thread::thread_self() );
 WyLock aa_toltab(wy_toltab.mutex());
 const Wy_Thread* optr=wy_toltab.find(ctid);
 if(optr==NULL) {
   //
   // Note: If the calling thread is not created by Wy_Thread,
   //       ::pthread_exit(r.c_errno()) is called.
   //
   ::pthread_exit((void*)exit_ret->c_repcode());
 }
 WyLock aa_optr(optr->_mtx);

 #ifndef NDEBUG
  if(optr->_thrd_stat!=Active) {
    // assertion: thread tmain() always sees Active state   
    WY_TERMINATE("");
  }
 #endif
 const_cast<Wy_Thread*>(optr)->_exit_obj.reset(exit_ret);
 ::pthread_exit( &const_cast<Wy_Thread*>(optr)->_exit_obj );
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const std::bad_alloc&) {
#ifdef WY_DEBUG
 Wy::cerr << "std::bad_alloc\n";
#endif
 WY_THROW( WyRet() );
}
catch(const std::exception&) {
#ifdef WY_DEBUG
 Wy::cerr << "std::exception\n";
#endif
 WY_THROW( WyRet() );
};

size_t Wy_Thread::active_threads(void)
try {
 WyRet r;
 if((r=wy_toltab.mutex().lock())!=OK) {
   WY_THROW(r);
 }
 size_t retv( wy_toltab.size() );
 if((r=wy_toltab.mutex().unlock())!=OK) {
   WY_THROW(r);
 }
 return(retv);
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

size_t Wy_Thread::max_threads(void) WY__TSPC()
{
 return WY__MAX_THRDS;
};
//----------------------------------------------------------------------------

/*
 note2: Wy_Thread assert the function call never fail:
        :::pthread_setcancelstate( PTHREAD_CANCEL_ENABLE: PTHREAD_CANCEL_DISABLE
                         , &oldstate)
        it might be better let this function always succeed.
 note5:
       ::pthread_cancel found to return EAGAIN (on RH Linux9.0)
       while man. indicates only fail on ESRCH.
 note6: Wy_Thread created threads always returns &_exit_obj, which is
        already set with desired value except cancelled.
        (So, value of PTHREAD_CANCELED cannot be any valid address, it is -1
         as I saw the source)
 note7: After successfully join _thrd_id is set to default
        (Wy_ThreadID::Default_ThreadID). see note8, program need to join
        a Stopped thread. This ensures join exactly once.
 note8: _thrd_stat transition general rule:
          Null -> Active    .. by begin()
          Active -> Stopped .. by _tmain_close()
          * -> Null         .. by reset()
        In Stopped state, _thrd_id is not default untill joined. Several
        members will implicitly join the thread
*/
