/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2005

   operators <<,>> for WyByteFlow
   Simulate std::cin/cout/cerr usage syntax
*/

#define WYLIB_SOURCE
#include "wy_uty.h"
#include <unistd.h>

WyRet Wy::wy__bwrite(WyByteFlow& bf, const char* buf, size_t blen)
{
 WyRet r;
 const char* wr_ptr( buf );
 size_t count( blen );
 for(;;) {
   size_t n;
   if((r=bf.write(wr_ptr,count,n))!=OK) {
     WY_RETURN(r);
   }
   if(n==count) {
     return(OK);
   }
   if(n>count) {
     WY_TERMINATE("");
   }
   wr_ptr+=n;
   count-=n;
 }
 // UNREACHABLE
};

WyByteFlow& operator <<(WyByteFlow& os, const WyStr& str)
{
 WyRet r( Wy::wy__bwrite(os,str.data(),str.size()) );
 if(r!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 return(os);
};

WyByteFlow& operator <<(WyByteFlow& os, WyCSeg str)
{
 WyRet r( Wy::wy__bwrite(os,str.begin(),str.size()) );
 if(r!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 return(os);
};

WyByteFlow& operator <<(WyByteFlow& os, const char* cstr)
{
 WyRet r( Wy::wy__bwrite(os,cstr,std::strlen(cstr)) );
 if(r!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 return(os);
};

WyByteFlow& operator <<(WyByteFlow& os, char* cstr)
{
 WyRet r( Wy::wy__bwrite(os,cstr,std::strlen(cstr)) );
 if(r!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 return(os);
};

WyByteFlow& operator <<(WyByteFlow& os, char ch)
{
 WyRet r( Wy::wy__bwrite(os,&ch,sizeof(ch)) );
 if(r!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 return(os);
};

WyByteFlow& operator <<(WyByteFlow& os, const void* p)
try {
 WyRet r;
 WyStr str(WyCSeg("0x",2));
 if((r=Wy::_catstr(str,(unsigned int)p,int(16)))!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 if((r=Wy::wy__bwrite(os,str.data(),str.size()))!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 return(os);
}
catch(const WyStr::Reply& e) {
 WY_THROW( WyByteFlow::Reply(e) );   // only Wym_ENOMEM possible
};

//-------------------------
template <typename NumT>
inline WyByteFlow& wy__put_value(WyByteFlow& bf, NumT num)
try {
 WyRet r;
 WyStr str;
 if((r=Wy::_mkstr(str,num,int(10)))!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 if((r=Wy::wy__bwrite(bf,str.data(),str.size()))!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 return(bf);
}
catch(const WyStr::Reply& e) {
 WY_THROW( WyByteFlow::Reply(e) );
};

WyByteFlow& operator <<(WyByteFlow& bf, signed char num)
{ return wy__put_value(bf,num); };
WyByteFlow& operator <<(WyByteFlow& bf, signed short num)
{ return wy__put_value(bf,num); };
WyByteFlow& operator <<(WyByteFlow& bf, signed int num)
{ return wy__put_value(bf,num); };
WyByteFlow& operator <<(WyByteFlow& bf, signed long num)
{ return wy__put_value(bf,num); };
WyByteFlow& operator <<(WyByteFlow& bf, signed long long num)
{ return wy__put_value(bf,num); };

WyByteFlow& operator <<(WyByteFlow& bf, unsigned char num)
{ return wy__put_value(bf,num); };
WyByteFlow& operator <<(WyByteFlow& bf, unsigned short num)
{ return wy__put_value(bf,num); };
WyByteFlow& operator <<(WyByteFlow& bf, unsigned int num)
{ return wy__put_value(bf,num); };
WyByteFlow& operator <<(WyByteFlow& bf, unsigned long num)
{ return wy__put_value(bf,num); };
WyByteFlow& operator <<(WyByteFlow& bf, unsigned long long num)
{ return wy__put_value(bf,num); };

WyByteFlow& operator <<(WyByteFlow& bf, float num)
{ return wy__put_value(bf,num); };
WyByteFlow& operator <<(WyByteFlow& bf, double num)
{ return wy__put_value(bf,num); };

//-------------------------
WyByteFlow& operator >>(WyByteFlow& bf, WyStr& str)
{
 str.reset();
 size_t n_rd;
 WyRet r( bf.read(str,str._capacity(),n_rd) );
 if(r!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 if(n_rd!=str.size()) {
   WY_THROW( WyByteFlow::Reply(Wym_EIO) );
 }
 return(bf);
};

WyByteFlow& operator >>(WyByteFlow& bf, char& ch)
{
 size_t n_rd;
 WyRet r( bf.read(&ch,sizeof(ch),n_rd) );
 if(r!=OK) {
   WY_THROW( WyByteFlow::Reply(r) );
 }
 if(n_rd!=sizeof(ch)) {
   if(n_rd==0) {
     WY_THROW( WyByteFlow::Reply(Wym_ENOENT) );
   }
   WY_THROW( WyByteFlow::Reply(Wym_EIO) );
 }
 return(bf);
};

/*
#include <wyiovec.h>
WyByteFlow& operator >>(WyByteFlow& bf, WyIOVec iov)
{
};*/

WyRet Wy::gethostname(WyStr& buf)
try {
#ifndef HOST_NAME_MAX      // man. indicated this name but not found
#define HOST_NAME_MAX 128
#endif
 WyRet r;
 WyStr tmp;
 if((r=tmp.resize(HOST_NAME_MAX))!=OK) {
   WY_RETURN(r);
 }
 const int v=::gethostname(&tmp[0],tmp.size()+1); // guess there is one extra
                                                  // byte at the end of data
 if(v!=0) {
   WY_RETURN(errno);
 }
 if((r=tmp.resize( ::strlen(tmp.data()) ))!=OK) {
   WY_THROW(r);
 }
 buf.swap(tmp);
 return(OK);
}
catch(const WyStr::Reply& e) {
  return(e);
};
