/* Copyright is licensed by GNU LGPL.                 by I.J.Wang 2004

   .One intent of WyCSeg is for checking null pointer once for all.
    Another is for simplying function argument, hence member function overloads
    for indicating char array.
   .Vagueness of class reponsibility.
   .Practices so far have shown WyCSeg is infrequently used in program
    The reason probably is that zero-terminated string is built-in.
    E.g. 
         WyStr f(const WyStr& x) {
           f( WyCSeg(i>0? "hello" : "goodbye"); );
           return( WyCSeg("ok") );
         };

      Or, 
         WyStr f(const WyStr& x) {
           static const WyCSeg C1("hello");
           static const WyCSeg C2("goodbye");
           static const WyCSeg C3("ok");
           f( i>0? C1 : C2; );
           return(C3);
         };

*/
#ifndef WYCSEG_H__
#define WYCSEG_H__
#define WYCSEG_VERSION 30

#include "wyret.h"
#include <cstring>
#include <string.h>   // for memmem,memrchr

class WyCSeg {
  public:
    typedef size_t size_type;
    typedef char value_type;
  private:
    inline WyCSeg(const value_type* p,size_type s1, size_type s2)
          : _pa(p), _pz(p+ ((s1<s2)? s1:s2))
          {};
    static const value_type* DefaultAddr;
    static const size_type DefaultSize=0;

    const value_type* _pa;  // begin of the segment
    const value_type* _pz;  // end of the segment (past-the-end value)

    // Hidden (vague)
    bool operator ==(const WyCSeg& rhs);
    bool operator !=(const WyCSeg& rhs);
    // lib rule implies operator= should also be hidden
    const WyCSeg& operator =(const WyCSeg& rhs) WY__TSPC();
    const WyCSeg& operator =(const value_type* cstr);

  public:
    WY_THROW_REPLY;

    WyCSeg() WY__TSPC()
          : _pa(DefaultAddr), _pz(DefaultAddr) {};

    WyCSeg(const WyCSeg& s) WY__TSPC()
          : _pa(s._pa), _pz(s._pz) {};

    // Note: the trailing zero is not counted for C-type string
    //
    WyCSeg(const value_type* cstr)
          : _pa(cstr),_pz(cstr)
          {
            if(cstr==NULL) {
              WY_THROW( Reply(Wym_EFAULT) );
            }
            _pz=cstr+std::strlen(cstr);
          };

    WyCSeg(const value_type* p,size_type s) WY__TSPC(Reply)
          : _pa(p), _pz(p+s)
          {
            if(p==NULL) {
              WY_THROW( Reply(Wym_EFAULT) ); 
            }
          };

    WyCSeg(const value_type* head, const value_type* tail) WY__TSPC(Reply)
          : _pa(head), _pz(tail)
          {
            if((head==NULL)||(tail==NULL)) {
              WY_THROW( Reply(Wym_EFAULT) );
            }
            if(ptrdiff_t(tail-head)<0) {
              WY_THROW( Reply(Wym_ERANGE) );
            }
          };

    WyCSeg(WyCSeg& s, Wy::ByMove_t) WY__TSPC()
          : _pa(s._pa), _pz(s._pz) {
            #ifndef NDEBUG
             s._pa=s._pz=NULL; 
            #endif
          };

    bool is_default(void) WY__TSPC()
          {
            // assert _pz==DefaultAddr;  
            return _pa==DefaultAddr;  
          };
    const value_type* begin(void) const WY__TSPC()
          { return(_pa); };
    const value_type* end(void) const WY__TSPC()
          { return _pz; };
    size_type size(void) const WY__TSPC()
          { return(_pz-_pa); };
    void reset() WY__TSPC()
          { _pa=DefaultAddr; _pz=DefaultAddr; };
    void reset(const WyCSeg& s) WY__TSPC()
          { _pa=s._pa; _pz=s._pz; };
    WyRet reset(const value_type* cstr)
          {
            if(cstr==NULL) {
              WY_RETURN(Wym_EFAULT);
            }
            _pa=cstr;
            _pz=cstr+std::strlen(cstr);
            return(OK);
          };
    WyRet reset(const value_type* p,size_type s) WY__TSPC()
          {
            if(p==NULL) {
              WY_RETURN(Wym_EFAULT);
            }
            _pa=p;
            _pz=p+s; 
            return(OK);
          };
    WyRet reset(const value_type* head, const value_type* tail) WY__TSPC()
          {
            if((head==NULL)||(tail==NULL)) {
              WY_RETURN( Wym_EFAULT );
            }
            if(tail<head) {
              WY_RETURN( Wym_ERANGE );
            }
            _pa=head;
            _pz=tail;
            return(OK);
          };

    void swap(WyCSeg& ano) WY__TSPC()
          {
            Wy__Base::vswap(_pa,ano._pa);
            Wy__Base::vswap(_pz,ano._pz);
          };

    const value_type& operator[](size_type idx) const
          { return _pa[idx]; };

    WyCSeg cseg(size_type idx) const WY__TSPC(Reply)
          {
            if(idx>static_cast<size_type>(_pz-_pa)) {
              WY_THROW( Reply(Wym_EINVAL) );
            }
            return WyCSeg(_pa+idx,_pz);
          };
    WyCSeg cseg(size_type idx, size_type s) const WY__TSPC(Reply)
          {
            const size_type seg_len(_pz-_pa);
            if(idx>seg_len) {
              WY_THROW( Reply(Wym_EINVAL) );
            }
            return WyCSeg(_pa+idx,seg_len-idx,s);
          };

        /*
            No 'find', for too much sementics involves the pointed object.
            use const char *Wy::_memmem(const WyCSeg&,const WyCSeg&)
        */
         
        // Note: C++ Standard looked to me that pointer arithmetic
        //       on array returns the difference of index. On my machine
        //       this is true. So this member is defined 
        //       (ptr comparison is used, but not found in C++ language standard)
        //
        // [Ret] true= there exist one element in common
        //      false= otherwise
        //
        bool is_overlap(const WyCSeg& s) const WY__TSPC()
              {
                if((_pa==_pz)||(s._pa==s._pz)) {
                  return(false);
                }
                if(_pa<s._pa) {
                  return(_pz>s._pa);
                } else if(_pa>s._pa) {
                  return(_pa<s._pz);
                } else {
                  return(true);
                }
              };
        };

namespace Wy {

 // Note: This function is in case source code really uses two pointers to save
 //       two explicit calls strlen(..).
 //       Functionality simulates Wy::strcmp(const WyCSeg&,const WyCSeg&)
 //
 inline int _strcmp(const char* s1, const char* s2) WY__TSPC(WyCSeg::Reply)
 {
   if((s1==NULL)||(s2==NULL)) {
     WY_THROW( WyCSeg::Reply(Wym_EFAULT) );
   }
   return std::strcmp(s1,s2); 
 };

 inline int wy_strcmp(const WyCSeg& s1, const WyCSeg& s2) WY__TSPC()
 {
  const WyCSeg::size_type Lft(s1.size());
  const WyCSeg::size_type Rit(s2.size());
  if(Lft!=Rit) {
    return Lft>Rit? 1:-1;
  }
  return std::memcmp(s1.begin(),s2.begin(),Rit);
 };

 inline int _strcmp(const WyCSeg& s1, const WyCSeg& s2) WY__TSPC()
 { return Wy::wy_strcmp(s1,s2); };

 inline int _strcmp(const WyCSeg& s1, const char* s2) WY__TSPC(WyCSeg::Reply)
 { return Wy::wy_strcmp(s1,WyCSeg(s2)); };

 inline int _strcmp(const char* s1, const WyCSeg& s2) WY__TSPC(WyCSeg::Reply)
 { return Wy::wy_strcmp(WyCSeg(s1),s2); };

 inline void _strcpy(char* dest, WyCSeg src)
 { std::memcpy(dest,src.begin(),src.size()); };

 inline void _strmove(char* dest, char* src, size_t n)
 { std::memmove(dest,src,n); };

 inline const char* _strchr(const WyCSeg& s, char c) WY__TSPC()
 { return static_cast<const char*>(std::memchr(s.begin(), c, s.size())); };

#ifdef _GNU_SOURCE
 inline const char* _strrchr(const WyCSeg& s, char c) WY__TSPC()
 { return static_cast<const char*>(::memrchr(s.begin(), c, s.size())); };

 inline const char* _strstr(const WyCSeg& hstk, const WyCSeg& ndl) WY__TSPC()
 { return static_cast<const char*>(::memmem(hstk.begin(),hstk.size(), ndl.begin(), ndl.size())); };
#endif

};

#endif   // End WYCSEG_H__
