/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003
*/

#define WYLIB_SOURCE
#include "wyret.h"
#include <cerrno>
#include <cstring>     // for strlen
#include <pthread.h>   // for mutex

//-------------------------------------------------
// Definition for Wy_RepInfo
//
// Note: Order of code is significant
//-------------------------------------------------
const Wy_RepInfo::RepCodeType Wy_RepInfo::_default_rcode=Wy::_default_errno;

Wy_RepInfo::Wy_RepInfo(const char* cstr) WY__NOTHROW__
 : _rcode(_default_rcode), _rstr(cstr)
{
 if(cstr==NULL) {
   WY_TERMINATE("");
 }
};

Wy_RepInfo::Wy_RepInfo(Wy_RepInfo::RepCodeType rc, const char* cstr) WY__NOTHROW__
 : _rcode(rc), _rstr(cstr)
{
 if(cstr==NULL) {
   WY_TERMINATE("");
 }
};

Wy_RepInfo::RepCodeType Wy_RepInfo::c_repcode(void) const WY__NOTHROW__
{
 return _rcode; 
};

const char* Wy_RepInfo::c_repstr(void) const WY__NOTHROW__
{
 return _rstr;
};

size_t Wy_RepInfo::get_reply(char* buf, size_t blen) const WY__NOTHROW__
{
 size_t slen( std::strlen(_rstr) );
 if(buf==NULL) {
   return slen;
 }
 if(blen<slen) {
   slen=blen;
 }
 std::memcpy(buf,_rstr,slen);
 return slen; 
};

//-------------------------------------------------
// Definition for WyRet
//
// Note: Order of code is significant
//-------------------------------------------------
class Wy__ErrNo : public Wy_RepInfo {
    const char* _ap_str;   // application reply string
    size_t      _ap_len;   // length of char's pointed by _ap_buf
    mutable ::pthread_mutex_t _mtx;

    // Hidden
    const Wy__ErrNo& operator =(const Wy__ErrNo&);
    bool operator ==(const Wy__ErrNo&);
    void* operator new(size_t);

  public:

    // [Syn] Construct object
    //
    //       c_repcode()  = en
    //       c_repstr()= cstr
    //     get_reply()= copy out cstr
    //
    Wy__ErrNo(Wy__ErrNoType en, const char* cstr) WY__NOTHROW__
       : Wy_RepInfo(en,cstr), _ap_str(cstr), _ap_len(std::strlen(cstr))
         {
           ::pthread_mutex_init(&_mtx,NULL);
         };

    ~Wy__ErrNo() WY__NOTHROW__
         {
           ::pthread_mutex_destroy(&_mtx); 
         };

    // Note: .Copy constructor is not defined.
    //       .Compiler (g++) seems need to 'see' this declaration.
    //        This may be from Standard requirement.
    //       .Wy_RepInfo object is not supposed to be derived by copy
    //        constructor
    //
    Wy__ErrNo(const Wy__ErrNo&) WY__NOTHROW__;

    // [Derived]
    // int c_repcode(void) const throw()
    // const char* c_repstr(void) const throw()

    // [Override]
    //
    size_t get_reply(char* buf, size_t blen) const WY__NOTHROW__
         {
           if(::pthread_mutex_lock(&_mtx)!=0) {
             WY_TERMINATE("");
           }
           if(buf==NULL) {
             const size_t rv(_ap_len);
             if(::pthread_mutex_unlock(&_mtx)!=0) {
               WY_TERMINATE("");
             }
             return(rv);
           }
           const size_t slen( (blen<_ap_len)? blen:_ap_len);
           std::memcpy(buf,_ap_str,slen);
           if(::pthread_mutex_unlock(&_mtx)!=0) {
             WY_TERMINATE("");
           }
           return slen; 
         };

    // [Override]
    //
    void set_reply(const char* str, size_t slen) WY__NOTHROW__
         {
           if(::pthread_mutex_lock(&_mtx)!=0) {
             WY_TERMINATE("");
           }
           _ap_str=str;
           _ap_len=slen;
           if(::pthread_mutex_unlock(&_mtx)!=0) {
             WY_TERMINATE("");
           }
         };
};

// The object for default WyReply to refer to
static Wy__ErrNo wye_default_repinfo(Wy::_default_errno,Wy::_default_repstr);

//-------------------------------------------------
// Definition for WyRet
//
// Note: Order of code is significant
//-------------------------------------------------

Wy_RepInfo* const WyReply::_default_ri_ptr=&wye_default_repinfo;

// Wy__ErrNo object for errno
//
static Wy__ErrNo wye_tab[Wy::_max_errno+1]= {
  Wy__ErrNo(  0,"Success"),
  Wy__ErrNo(  1,"Operation not permitted"),
  Wy__ErrNo(  2,"No such file or directory"),
  Wy__ErrNo(  3,"No such process"),
  Wy__ErrNo(  4,"Interrupted system call"),
  Wy__ErrNo(  5,"Input/output error"),
  Wy__ErrNo(  6,"No such device or address"),
  Wy__ErrNo(  7,"Argument list too long"),
  Wy__ErrNo(  8,"Exec format error"),
  Wy__ErrNo(  9,"Bad file descriptor"),
  Wy__ErrNo( 10,"No child processes"),
  Wy__ErrNo( 11,"Resource temporarily unavailable"),
  Wy__ErrNo( 12,"Cannot allocate memory"),
  Wy__ErrNo( 13,"Permission denied"),
  Wy__ErrNo( 14,"Bad address"),
  Wy__ErrNo( 15,"Block device required"),
  Wy__ErrNo( 16,"Device or resource busy"),
  Wy__ErrNo( 17,"File exists"),
  Wy__ErrNo( 18,"Invalid cross-device link"),
  Wy__ErrNo( 19,"No such device"),
  Wy__ErrNo( 20,"Not a directory"),
  Wy__ErrNo( 21,"Is a directory"),
  Wy__ErrNo( 22,"Invalid argument"),
  Wy__ErrNo( 23,"Too many open files in system"),
  Wy__ErrNo( 24,"Too many open files"),
  Wy__ErrNo( 25,"Inappropriate ioctl for device"),
  Wy__ErrNo( 26,"Text file busy"),
  Wy__ErrNo( 27,"File too large"),
  Wy__ErrNo( 28,"No space left on device"),
  Wy__ErrNo( 29,"Illegal seek"),
  Wy__ErrNo( 30,"Read-only file system"),
  Wy__ErrNo( 31,"Too many links"),
  Wy__ErrNo( 32,"Broken pipe"),
  Wy__ErrNo( 33,"Numerical argument out of domain"),
  Wy__ErrNo( 34,"Numerical result out of range"),
  Wy__ErrNo( 35,"Resource deadlock would occur"),
  Wy__ErrNo( 36,"File name too long"),
  Wy__ErrNo( 37,"No locks available"),
  Wy__ErrNo( 38,"Function not implemented"),
  Wy__ErrNo( 39,"Directory not empty"),
  Wy__ErrNo( 40,"Too many levels of symbolic links"),
  Wy__ErrNo( 41,"Unknown error 41"), 
  Wy__ErrNo( 42,"No message of desired type"),
  Wy__ErrNo( 43,"Identifier removed"),
  Wy__ErrNo( 44,"Channel number out of range"),
  Wy__ErrNo( 45,"Level 2 not synchronized"),
  Wy__ErrNo( 46,"Level 3 halted"),
  Wy__ErrNo( 47,"Level 3 reset"),
  Wy__ErrNo( 48,"Link number out of range"),
  Wy__ErrNo( 49,"Protocol driver not attached"),
  Wy__ErrNo( 50,"No CSI structure available"),
  Wy__ErrNo( 51,"Level 2 halted"),
  Wy__ErrNo( 52,"Invalid exchange"),
  Wy__ErrNo( 53,"Invalid request descriptor"),
  Wy__ErrNo( 54,"Exchange full"),
  Wy__ErrNo( 55,"No anode"),
  Wy__ErrNo( 56,"Invalid request code"),
  Wy__ErrNo( 57,"Invalid slot"),
  Wy__ErrNo( 58,"Unknown error 58"),
  Wy__ErrNo( 59,"Bad font file format"),
  Wy__ErrNo( 60,"Device not a stream"),
  Wy__ErrNo( 61,"No data available"),
  Wy__ErrNo( 62,"Timer expired"),
  Wy__ErrNo( 63,"Out of streams resources"),
  Wy__ErrNo( 64,"Machine is not on the network"),
  Wy__ErrNo( 65,"Package not installed"),
  Wy__ErrNo( 66,"Object is remote"),
  Wy__ErrNo( 67,"Link has been severed"),
  Wy__ErrNo( 68,"Advertise error"),
  Wy__ErrNo( 69,"Srmount error"),
  Wy__ErrNo( 70,"Communication error on send"),
  Wy__ErrNo( 71,"Protocol error"),
  Wy__ErrNo( 72,"Multihop attempted"),
  Wy__ErrNo( 73,"RFS specific error"),
  Wy__ErrNo( 74,"Bad message"),
  Wy__ErrNo( 75,"Value too large for defined data type"),
  Wy__ErrNo( 76,"Name not unique on network"),
  Wy__ErrNo( 77,"File descriptor in bad state"),
  Wy__ErrNo( 78,"Remote address changed"),
  Wy__ErrNo( 79,"Can not access a needed shared library"),
  Wy__ErrNo( 80,"Accessing a corrupted shared library"),
  Wy__ErrNo( 81,".lib section in a.out corrupted"),
  Wy__ErrNo( 82,"Attempting to link in too many shared libraries"),
  Wy__ErrNo( 83,"Cannot exec a shared library directly"),
  Wy__ErrNo( 84,"Illegal byte sequence"),
  Wy__ErrNo( 85,"Interrupted system call should be restarted"),
  Wy__ErrNo( 86,"Streams pipe error"),
  Wy__ErrNo( 87,"Too many users"),
  Wy__ErrNo( 88,"Socket operation on non-socket"),
  Wy__ErrNo( 89,"Destination address required"),
  Wy__ErrNo( 90,"Message too long"),
  Wy__ErrNo( 91,"Protocol wrong type for socket"),
  Wy__ErrNo( 92,"Protocol not available"),
  Wy__ErrNo( 93,"Protocol not supported"),
  Wy__ErrNo( 94,"Socket type not supported"),
  Wy__ErrNo( 95,"Operation not supported"),
  Wy__ErrNo( 96,"Protocol family not supported"),
  Wy__ErrNo( 97,"Address family not supported by protocol"),
  Wy__ErrNo( 98,"Address already in use"),
  Wy__ErrNo( 99,"Cannot assign requested address"),
  Wy__ErrNo(100,"Network is down"),
  Wy__ErrNo(101,"Network is unreachable"),
  Wy__ErrNo(102,"Network dropped connection on reset"),
  Wy__ErrNo(103,"Software caused connection abort"),
  Wy__ErrNo(104,"Connection reset by peer"),
  Wy__ErrNo(105,"No buffer space available"),
  Wy__ErrNo(106,"Transport endpoint is already connected"),
  Wy__ErrNo(107,"Transport endpoint is not connected"),
  Wy__ErrNo(108,"Cannot send after transport endpoint shutdown"),
  Wy__ErrNo(109,"Too many references: cannot splice"),
  Wy__ErrNo(110,"Connection timed out"),
  Wy__ErrNo(111,"Connection refused"),
  Wy__ErrNo(112,"Host is down"),
  Wy__ErrNo(113,"No route to host"),
  Wy__ErrNo(114,"Operation already in progress"),
  Wy__ErrNo(115,"Operation now in progress"),
  Wy__ErrNo(116,"Stale NFS file handle"),
  Wy__ErrNo(117,"Structure needs cleaning"),
  Wy__ErrNo(118,"Not a XENIX named type file"),
  Wy__ErrNo(119,"No XENIX semaphores available"),
  Wy__ErrNo(120,"Is a named type file"),
  Wy__ErrNo(121,"Remote I/O error"),
  Wy__ErrNo(122,"Disk quota exceeded"),
  Wy__ErrNo(123,"No medium found"),
  Wy__ErrNo(124,"Wrong medium type"),
  Wy__ErrNo(125,"Operation canceled"),
  Wy__ErrNo(126,"Unknown error 126"),
  Wy__ErrNo(127,"Unknown error 127")
};

// libwy added errinfo objects 
//
static Wy__ErrNo wye_ETHRDCAN(Wy::_default_errno,"Thread canceled");
static Wy__ErrNo wye_EDIVZERO(Wy::_default_errno,"Divisor is zero");
static Wy__ErrNo wye_EMATHNEG(Wy::_default_errno,"Math negation failed");
static Wy__ErrNo wye_EBADTYPE(Wy::_default_errno,"Bad argument type");

//-------------------------------------------------------
// Note: Order of code is significant
//-------------------------------------------------------

// WyReply constants. Text is roughly the same as for C-library errno
//
 const WyReply OK( wye_tab[0] );
 const WyReply Wym_EPERM( wye_tab[1] );
 const WyReply Wym_ENOENT( wye_tab[2] );
 const WyReply Wym_ESRCH( wye_tab[3] );
 const WyReply Wym_EINTR( wye_tab[4] );
 const WyReply Wym_EIO	( wye_tab[5] );
 const WyReply Wym_ENXIO( wye_tab[6] );
 const WyReply Wym_E2BIG( wye_tab[7] );
 const WyReply Wym_ENOEXEC( wye_tab[8] );
 const WyReply Wym_EBADF( wye_tab[9] );
 const WyReply Wym_ECHILD( wye_tab[10] );
 const WyReply Wym_EAGAIN( wye_tab[11] );
 const WyReply Wym_ENOMEM( wye_tab[12] );
 const WyReply Wym_EACCES( wye_tab[13] );
 const WyReply Wym_EFAULT( wye_tab[14] );
 const WyReply Wym_ENOTBLK( wye_tab[15] );
 const WyReply Wym_EBUSY( wye_tab[16] );
 const WyReply Wym_EEXIST( wye_tab[17] );
 const WyReply Wym_EXDEV( wye_tab[18] );
 const WyReply Wym_ENODEV( wye_tab[19] );
 const WyReply Wym_ENOTDIR( wye_tab[20] );
 const WyReply Wym_EISDIR( wye_tab[21] );
 const WyReply Wym_EINVAL( wye_tab[22] );
 const WyReply Wym_ENFILE( wye_tab[23] );
 const WyReply Wym_EMFILE( wye_tab[24] );
 const WyReply Wym_ENOTTY( wye_tab[25] );
 const WyReply Wym_ETXTBSY( wye_tab[26] );
 const WyReply Wym_EFBIG( wye_tab[27] );
 const WyReply Wym_ENOSPC( wye_tab[28] );
 const WyReply Wym_ESPIPE( wye_tab[29] );
 const WyReply Wym_EROFS( wye_tab[30] );
 const WyReply Wym_EMLINK( wye_tab[31] );
 const WyReply Wym_EPIPE( wye_tab[32] );
 const WyReply Wym_EDOM( wye_tab[33] );
 const WyReply Wym_ERANGE( wye_tab[34] );
 const WyReply Wym_EDEADLK( wye_tab[35] );
 const WyReply Wym_ENAMETOOLONG( wye_tab[36] );
 const WyReply Wym_ENOLCK( wye_tab[37] );
 const WyReply Wym_ENOSYS( wye_tab[38] );
 const WyReply Wym_ENOTEMPTY( wye_tab[39] );
 const WyReply Wym_ELOOP( wye_tab[40] );
 const WyReply Wym_EWOULDBLOCK( wye_tab[11] );   // ==EAGAIN
 const WyReply Wym_E41( wye_tab[41] );
 const WyReply Wym_ENOMSG( wye_tab[42] );
 const WyReply Wym_EIDRM( wye_tab[43] );
 const WyReply Wym_ECHRNG( wye_tab[44] );
 const WyReply Wym_EL2NSYNC( wye_tab[45] );
 const WyReply Wym_EL3HLT( wye_tab[46] );
 const WyReply Wym_EL3RST( wye_tab[47] );
 const WyReply Wym_ELNRNG( wye_tab[48] );
 const WyReply Wym_EUNATCH( wye_tab[49] );
 const WyReply Wym_ENOCSI( wye_tab[50] );
 const WyReply Wym_EL2HLT( wye_tab[51] );
 const WyReply Wym_EBADE( wye_tab[52] );
 const WyReply Wym_EBADR( wye_tab[53] );
 const WyReply Wym_EXFULL( wye_tab[54] );
 const WyReply Wym_ENOANO( wye_tab[55] );
 const WyReply Wym_EBADRQC( wye_tab[56] );
 const WyReply Wym_EBADSLT( wye_tab[57] );
 const WyReply Wym_EDEADLOCK( wye_tab[35] );     // ==EDEADLK
 const WyReply Wym_E58( wye_tab[58] );
 const WyReply Wym_EBFONT( wye_tab[59] );
 const WyReply Wym_ENOSTR( wye_tab[60] );
 const WyReply Wym_ENODATA( wye_tab[61] );
 const WyReply Wym_ETIME( wye_tab[62] );
 const WyReply Wym_ENOSR( wye_tab[63] );
 const WyReply Wym_ENONET( wye_tab[64] );
 const WyReply Wym_ENOPKG( wye_tab[65] );
 const WyReply Wym_EREMOTE( wye_tab[66] );
 const WyReply Wym_ENOLINK( wye_tab[67] );
 const WyReply Wym_EADV( wye_tab[68] );
 const WyReply Wym_ESRMNT( wye_tab[69] );
 const WyReply Wym_ECOMM( wye_tab[70] );
 const WyReply Wym_EPROTO( wye_tab[71] );
 const WyReply Wym_EMULTIHOP( wye_tab[72] );
 const WyReply Wym_EDOTDOT( wye_tab[73] );
 const WyReply Wym_EBADMSG( wye_tab[74] );
 const WyReply Wym_EOVERFLOW( wye_tab[75] );
 const WyReply Wym_ENOTUNIQ( wye_tab[76] );
 const WyReply Wym_EBADFD( wye_tab[77] );
 const WyReply Wym_EREMCHG( wye_tab[78] );
 const WyReply Wym_ELIBACC( wye_tab[79] );
 const WyReply Wym_ELIBBAD( wye_tab[80] );
 const WyReply Wym_ELIBSCN( wye_tab[81] );
 const WyReply Wym_ELIBMAX( wye_tab[82] );
 const WyReply Wym_ELIBEXEC( wye_tab[83] );
 const WyReply Wym_EILSEQ( wye_tab[84] );
 const WyReply Wym_ERESTART( wye_tab[85] );
 const WyReply Wym_ESTRPIPE( wye_tab[86] );
 const WyReply Wym_EUSERS( wye_tab[87] );
 const WyReply Wym_ENOTSOCK( wye_tab[88] );
 const WyReply Wym_EDESTADDRREQ( wye_tab[89] );
 const WyReply Wym_EMSGSIZE( wye_tab[90] );
 const WyReply Wym_EPROTOTYPE( wye_tab[91] );
 const WyReply Wym_ENOPROTOOPT( wye_tab[92] );
 const WyReply Wym_EPROTONOSUPPORT( wye_tab[93] );
 const WyReply Wym_ESOCKTNOSUPPORT( wye_tab[94] );
 const WyReply Wym_EOPNOTSUPP( wye_tab[95] );
 const WyReply Wym_EPFNOSUPPORT( wye_tab[96] );
 const WyReply Wym_EAFNOSUPPORT( wye_tab[97] );
 const WyReply Wym_EADDRINUSE( wye_tab[98] );
 const WyReply Wym_EADDRNOTAVAIL( wye_tab[99] );
 const WyReply Wym_ENETDOWN( wye_tab[100] );
 const WyReply Wym_ENETUNREACH( wye_tab[101] );
 const WyReply Wym_ENETRESET( wye_tab[102] );
 const WyReply Wym_ECONNABORTED( wye_tab[103] );
 const WyReply Wym_ECONNRESET( wye_tab[104] );
 const WyReply Wym_ENOBUFS( wye_tab[105] );
 const WyReply Wym_EISCONN( wye_tab[106] );
 const WyReply Wym_ENOTCONN( wye_tab[107] );
 const WyReply Wym_ESHUTDOWN( wye_tab[108] );
 const WyReply Wym_ETOOMANYREFS( wye_tab[109] );
 const WyReply Wym_ETIMEDOUT( wye_tab[110] );
 const WyReply Wym_ECONNREFUSED( wye_tab[111] );
 const WyReply Wym_EHOSTDOWN( wye_tab[112] );
 const WyReply Wym_EHOSTUNREACH( wye_tab[113] );
 const WyReply Wym_EALREADY( wye_tab[114] );
 const WyReply Wym_EINPROGRESS( wye_tab[115] );
 const WyReply Wym_ESTALE( wye_tab[116] );
 const WyReply Wym_EUCLEAN( wye_tab[117] );
 const WyReply Wym_ENOTNAM( wye_tab[118] );
 const WyReply Wym_ENAVAIL( wye_tab[119] );
 const WyReply Wym_EISNAM( wye_tab[120] );
 const WyReply Wym_EREMOTEIO( wye_tab[121] );
 const WyReply Wym_EDQUOT( wye_tab[122] );
 const WyReply Wym_ENOMEDIUM( wye_tab[123] );
 const WyReply Wym_EMEDIUMTYPE( wye_tab[124] );
 const WyReply Wym_E125( wye_tab[125] );
 const WyReply Wym_E126( wye_tab[126] );
 const WyReply Wym_E127( wye_tab[127] );

// libwy added global reply constants
//
 const WyReply Wym_ETHRDCAN( wye_ETHRDCAN );
 const WyReply Wym_EDIVZERO( wye_EDIVZERO );
 const WyReply Wym_EMATHNEG( wye_EMATHNEG );
 const WyReply Wym_EBADTYPE( wye_EBADTYPE );

//-----------------------------------------------------------------------------
// Definition code for Wy_SrcLoc
//
// Note: Order of code is significant
//-----------------------------------------------------------------------------
// No code yet

//-----------------------------------------------------------------------------
// Definition code for WyReply
//
// Note: Order of code is significant
//-----------------------------------------------------------------------------
Wy_RepInfo* WyReply::_en2ptr(Wy__ErrNoType en) WY__NOTHROW__
{
 if((en<0)||(en>Wy::_max_errno)) {
   return(_default_ri_ptr);
 }
#ifdef WY_DEBUG
 if(wye_tab[en].c_repcode()!=en) {
   WY_TERMINATE("");    // assertion failed
 }
#endif
 return(&wye_tab[en]);
};

Wy__ErrNoType WyReply::reset(Wy__ErrNoType en) WY__NOTHROW__
{
 if((en<0)||(en>Wy::_max_errno)) {
   _ri_ptr=_default_ri_ptr;
   return(Wy::_default_errno);
 }
 _ri_ptr=&wye_tab[en];
#ifdef WY_DEBUG
 if(_ri_ptr->c_repcode()!=en) {
   WY_TERMINATE("");    // assertion failed
 }
#endif
 return(en);
};

//-------------------------------------------------
// Definition for WyRet
//
// Note: Order of code is significant
//-------------------------------------------------
// No code yet

//-------------------------------------------------
// Warning: Wy_ErrInfo is obsolete, will be removed in latter version
//
// Definition for Wy_ErrInfo
//-------------------------------------------------
#ifndef WY_DEBUG
size_t Wy_ErrInfo::get_reply(char* buf, size_t blen) const
{
 if(buf==NULL) {
   return (_ustr==NULL)? Wy_RepInfo::get_reply(buf,blen): _ulen;
 }
 if(_ustr==NULL) {
   return Wy_RepInfo::get_reply(buf,blen);
 }
 const size_t clen( _ulen<blen? _ulen:blen);
 std::memcpy(buf,_ustr,clen);
 return(clen);
};

void Wy_ErrInfo::set_reply(const char* buf, size_t blen)
{
 _ustr=buf;
 _ulen=blen;
};
#endif
