/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2005
*/

#define WYLIB_SOURCE
#include "wysysfile.h"
#include "wy_atdestroy.h"
#include <typeinfo>
#include <memory>

const char WySysFile::class_name[]="WySysFile";

#ifdef WY_DEBUG
int WySysFile::_obj_cnt=0;
#endif

//-----------------------------------------------------------------
Wy__FTSet::Wy__FTSet() WY__NOTHROW__
 : _bit(0) 
{};

Wy__FTSet::Wy__FTSet(const Wy__FTSet& fts) WY__NOTHROW__
 : _bit(fts._bit)
{};

Wy__FTSet::Wy__FTSet(mode_t m)
{
 switch(m) {
   case S_IFREG:  _bit=0x01; break;
   case S_IFCHR:  _bit=0x02; break;
   case S_IFSOCK: _bit=0x04; break;
   case S_IFIFO:  _bit=0x08; break;
   case S_IFDIR:  _bit=0x10; break;
   case S_IFLNK:  _bit=0x20; break;
   case S_IFBLK:  _bit=0x40; break;
   default:
     WY_THROW( WyRet() );
 };
};

Wy__FTSet Wy__FTSet::operator |(const Wy__FTSet& rhs) const WY__NOTHROW__
{
 Wy__FTSet res(*this);
 res._bit|=rhs._bit;
 return(res);
};

bool Wy__FTSet::contain(mode_t ft) const WY__NOTHROW__
{
 Wy__FTSet tmp(ft);
 return _bit&tmp._bit;
};

bool Wy__FTSet::is_joint(const Wy__FTSet& fts) const WY__NOTHROW__
{
 return _bit&fts._bit;
};

//-----------------------------------------------
Wy__CIF::Wy__CIF(const Wy__FTSet& fts) WY__NOTHROW__
 : _ftset(fts) {
};

bool Wy__CIF::has_ftype(mode_t ft) const WY__NOTHROW__
{
 return _ftset.contain(ft);
};

bool Wy__CIF::has_common_ftype(const Wy__CIF& cif) const WY__NOTHROW__
{
 return _ftset.is_joint(cif._ftset);
};

//-------------------------------------------------------------------------

// [Cancel Point]
// [Internal] close fd (modified version of ::close)
//
// Note: This function assumes fd is always sucessfully closed (freely reusable)
//       fd is also closed in failure or cancellation
//
//     1. Test program t_close.cpp has never caught EINTR from ::close(int)
//        (Fedora Core 3)
//
//     2. A post 37748e5c16749481_002.html suggests that the file descriptor
//        has been closed by ::close(int) returning EINTR on Linux
//
//     3. Spec. from these URL's indicate that close(fd) failure postcondition
//        is unspecified (except EBADF), but this is not deterministic for this
//        concern.
//        http://www.die.net/doc/linux/man/man3/close.3.html
//        http://www.opengroup.org/onlinepubs/000095399/functions/close.html
//
// [Ret] errno
//
int WySysFile::wy_close(const Wy__TypeFD& fd)
{
#ifdef WY_DEBUG
 if(fd!=WY__DEFAULT_FD) {
   --wyc__fdcnt;
 }
#endif
 return (::close(fd)==0)? 0: errno;
};

const Wy__CIF WySysFile::SysCif(
        Wy__FTSet(S_IFREG)|
        Wy__FTSet(S_IFCHR)|
        Wy__FTSet(S_IFSOCK)| 
        Wy__FTSet(S_IFIFO)|
        Wy__FTSet(S_IFDIR)|
        Wy__FTSet(S_IFLNK)|
        Wy__FTSet(S_IFBLK)
      );

inline void WySysFile::_init(const char* pathname,int flags)
{
 if(flags&O_CREAT) {
   WY_THROW( Reply(Wym_EINVAL) );
 }
 _m_fd=wyc_open(pathname,flags);
 if(_m_fd==-1) {
   WY_THROW( Reply(WyReply(errno)) );
 }
 Wy_AtDestroy<int,const Wy__TypeFD&> rrid(WySysFile::wy_close,_m_fd);
 WyFileStat stt;
 if(wyc_fstat(_m_fd,stt.wy_stat_ptr())==-1) {
   WY_THROW( Reply(WyReply(errno)) );
 }
 _m_ftype=stt.mode()&S_IFMT;
 if((_cif.has_ftype(_m_ftype))==false) {
   WY_THROW( Reply(Wym_EBADF) );
 }
 rrid.release();
};

WySysFile::WySysFile(const Wy__CIF& cif)
 : _m_fd(DefaultFd),_m_ftype(DefaultFType),_cif(cif)
{
#ifdef WY_DEBUG
 ++_obj_cnt;
#endif
};

WySysFile::WySysFile(WyFileHandle fh, const Wy__CIF& cif)
 : _m_fd(DefaultFd),_m_ftype(DefaultFType),_cif(cif)
{
 if(fh.is_default()) {
#ifdef WY_DEBUG
 ++_obj_cnt;
#endif
   return;
 }
 WyRet r;
 WyFileStat stt;
 if((r=fh.stat(stt))!=OK) {
   WY_THROW( Reply(r) );
 }
 _m_ftype=stt.mode()&S_IFMT;
 if((_cif.has_ftype(_m_ftype))==false) {
   WY_THROW( Reply(Wym_EBADF) );
 }
 _m_fd=wyc_dup(fh.fd());
 if(_m_fd==-1) {
   WY_THROW( Reply(WyReply(errno)) );
 }
#ifdef WY_DEBUG
 ++_obj_cnt;
#endif
};

WySysFile::WySysFile(const WySysFile& sysf, const Wy__CIF& cif)
 : _m_fd(DefaultFd),_m_ftype(sysf._m_ftype),_cif(cif)
{
 if(sysf.is_default()) {
   // default object is runtime checked for types that may be consistent 
   if(_cif.has_common_ftype(sysf._cif)==false) {
     WY_THROW( Reply(Wym_EBADF) );
   }
#ifdef WY_DEBUG
 ++_obj_cnt;
#endif
   return;
 }
 if(_cif.has_ftype(_m_ftype)==false) {
   WY_THROW( Reply(Wym_EBADF) );
 }
 _m_fd=wyc_dup(sysf._m_fd);
 if(_m_fd==-1) {
   WY_THROW( Reply(WyReply(errno)) );
 }
#ifdef WY_DEBUG
 ++_obj_cnt;
#endif
};

WySysFile::WySysFile(const char* pathname,int flags,const Wy__CIF& cif)
 : _m_fd(DefaultFd),_m_ftype(DefaultFType),_cif(cif)
{
 this->_init(pathname,flags);
#ifdef WY_DEBUG
 ++_obj_cnt;
#endif
};

WySysFile::WySysFile(const WyStr& pathname,int flags,const Wy__CIF& cif)
 : _m_fd(DefaultFd),_m_ftype(DefaultFType),_cif(cif)
{
 this->_init(pathname.c_str(),flags);
#ifdef WY_DEBUG
 ++_obj_cnt;
#endif
};

WySysFile::WySysFile(WySysFile& sysf,Wy::ByMove_t) WY__TSPC()
 : _m_fd(sysf._m_fd),_m_ftype(sysf._m_ftype),_cif(sysf._cif)
{
 // 'Clear' resource mark of sysf
 // This is not required but for safety
 sysf._m_fd=DefaultFd;
 sysf._m_ftype=DefaultFType;
};

WySysFile::~WySysFile()
try {
#ifdef WY_DEBUG
 --_obj_cnt;
#endif
 if(_m_fd==DefaultFd) {
   return;
 }
 Wy__Base::NoCancel noc;    // dtor cannot be cancellation point
 WySysFile::wy_close(_m_fd);  // return code ignored
}
catch(int e) {
 WY_THROW( WyRet(WyReply(e)) );
};

bool WySysFile::is_default(void) const WY__TSPC()
{
 return(_m_fd==DefaultFd);
};

WyFileHandle WySysFile::fh(void) const WY__TSPC()
{
 return WyFileHandle(_m_fd);
};

mode_t WySysFile::ftype(void) const WY__TSPC()
{ return _m_ftype; 
};

WyRet WySysFile::stat(WyFileStat& stt) const
{
 if(wyc_fstat(_m_fd,stt.wy_stat_ptr())==-1) {
   WY_RETURN(errno);
 }
 return(OK);
};

WyRet WySysFile::reset(void)
{
 if(_m_fd==DefaultFd) {
   #ifdef WY_DEBUG
     if(_m_ftype!=DefaultFType) {
       WY_THROW( WyRet() );
     }
   #endif
   return(OK);
 }
 #ifdef WY_DEBUG
   if(_m_ftype==DefaultFType) {
     WY_THROW( WyRet() );
   }
 #endif
 const Wy__TypeFD pfd(_m_fd);
 _m_fd=DefaultFd;
 _m_ftype=DefaultFType;
 const int v( WySysFile::wy_close(pfd) );
 if(v!=0) {
   WY_RETURN(WyReply(v));
 }
 return(OK);
};

// Failure poscondition for the following _reset(..) members:
//
//   return- object intact
//   throw - Reply: same as return
//           Others: complete or intact, which is unspecified
//   cancel- complete or intact, which is unspecified
//
WyRet WySysFile::reset(WyFileHandle fh)
{
 if(fh.is_default()) {
   this->reset();
   return(OK);
 }
 WyRet r;
 WyFileStat stt;
 if((r=fh.stat(stt))!=OK) {
   WY_RETURN(r);
 }
 const mode_t fh_ftype=stt.mode()&S_IFMT;
 if(_cif.has_ftype(fh_ftype)==false) {
   WY_RETURN(Wym_EBADF);
 }
 Wy__TypeFD tfd=wyc_dup(fh.fd());
 if(tfd==-1) {
   WY_RETURN(errno);
 }
 Wy__Base::vswap(_m_fd,tfd);
 _m_ftype=fh_ftype;
 WySysFile::wy_close(tfd);
 return(OK);
};

WyRet WySysFile::reset(const WySysFile& sysf)
{
 if(sysf.is_default()) {
   if(_cif.has_common_ftype(sysf._cif)==false) {
     WY_RETURN(Wym_EBADF);
   }
   this->reset();
   return(OK);
 }
 if(_cif.has_ftype(sysf._m_ftype)==false) {
   WY_RETURN(Wym_EBADF);
 }
 Wy__TypeFD tfd=wyc_dup(sysf._m_fd);
 if(tfd==-1) {
   WY_RETURN(errno);
 }
 Wy__Base::vswap(_m_fd,tfd);
 _m_ftype=sysf._m_ftype;
 WySysFile::wy_close(tfd);
 return(OK);
};

WyRet WySysFile::reset(const char*pathname, int flags)
{
 // note: report of closing provious fd is ignored
 //
 if(flags&O_CREAT) {
   WY_RETURN(Wym_EINVAL);
 }
 Wy__TypeFD tfd=wyc_open(pathname,flags);
 if(tfd==-1) {
   WY_RETURN(errno);
 }
 Wy_AtDestroy<int,const Wy__TypeFD&> rrid(WySysFile::wy_close,tfd);
 WyFileStat stt;
 if(wyc_fstat(tfd,stt.wy_stat_ptr())==-1) {
   WY_RETURN(errno);
 }
 const mode_t p_ftype=stt.mode()&S_IFMT;
 if(_cif.has_ftype(p_ftype)==false) {
   WY_RETURN(Wym_EBADF);
 }
 Wy__Base::vswap(_m_fd,tfd);
 _m_ftype=p_ftype;
 return(OK);
};

WyRet WySysFile::reset(const WyStr& pathname, int flags)
{
 return this->reset(pathname.c_str(),flags);
};

WyRet WySysFile::_swap(WySysFile& sysf) WY__TSPC()
{
 if(WY__EQU_TYPEID(*this,sysf)==false) {
   WY_RETURN(Wym_EBADTYPE);
 }
 Wy__Base::vswap(_m_fd,sysf._m_fd);
 Wy__Base::vswap(_m_ftype,sysf._m_ftype);
 return(OK);
};

WySysFile* WySysFile::_alloc(WyRet& r) const
try {
 if(WY__EQU_TYPEID(*this,WySysFile)==false) {
   r=Wym_ENOSYS; WY_HERE(r);        // _alloc not overridden
   return(NULL);
 }
 WySysFile* p=new(std::nothrow) WySysFile();
 if(p==NULL) {
   r=Wym_ENOMEM; WY_HERE(r);
 } else {
   r=OK;
 }
 return(p);
}
catch(const Reply& e) {
 r=e; WY_HERE(r);
 return(NULL);
}
catch(const WyRet& e) {
 WY_THROW( WyRet(e) );
};

void WySysFile::_overwrite(Wy__TypeFD fd, mode_t ftype) WY__NOTHROW__
{
 this->_m_fd=fd;
 this->_m_ftype=ftype;
};
