/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2005

   Documentation in file wysysfile.3wy
*/
#ifndef WYSYSFILE_H__
#define WYSYSFILE_H__
#define WYSYSFILE_VERSION 30

#include "wystr.h"
#include "wyfilehandle.h"
#include "wy__cfunc.h"   // for Wy::_delete..
#include <fcntl.h>       // for symbols (e.g. O_RDONLY..)
#include <sys/types.h>   // for symbols
#include <sys/stat.h>    // manual says (mode_t probably)

// [Syn] Set of file type
//
class Wy__FTSet {
    int _bit;
  public:
    Wy__FTSet() WY__NOTHROW__;
    Wy__FTSet(const Wy__FTSet&) WY__NOTHROW__;
    
    // [Syn] Construct object to contain one single element m
    //       m has to be one of the symbol: 
    //       S_IFREG,S_IFCHR,S_IFSOCK,S_IFIFO,S_IFDIR,S_IFLNK,S_IFBLK
    //
    Wy__FTSet(mode_t m);

    // [Syn] OR set operation
    //
    // [Ret] Wy__FTSet containing elements of *this and rhs
    //
    Wy__FTSet operator |(const Wy__FTSet& rhs) const WY__NOTHROW__;

    // [Syn] Is ft in *this
    //
    // [Ret] true = *this contains ft
    //       false= otherwise
    //
    bool contain(mode_t ft) const WY__NOTHROW__;

    // [Syn] Have *this and fts any element in common?
    //
    // [Ret] true = *this and fts have elements in common
    //       false= otherwise
    //
    bool is_joint(const Wy__FTSet& fts) const WY__NOTHROW__;
};

// [Syn] Class information of specific WySysFile's class family
//       This class is mostly for runtime check polymorphism
//
class Wy__CIF {
    Wy__FTSet _ftset;

    Wy__CIF() WY__NOTHROW__;
    Wy__CIF(const Wy__CIF&) WY__NOTHROW__;
  public:
    Wy__CIF(const Wy__FTSet&) WY__NOTHROW__;
    virtual ~Wy__CIF() WY__NOTHROW__{};

    // [Syn] *this contains the file type ft
    //
    // [Ret] true = *this contains the file type ft
    //       false= otherwise
    //
    bool has_ftype(mode_t ft) const WY__NOTHROW__;

    // [Syn] *this and argement Wy__CIF object has common file type
    //
    // [Ret] true = *this and argument cif has file type in common
    //       false= otherwise
    //
    bool has_common_ftype(const Wy__CIF& cif) const WY__NOTHROW__;
};

class WySysFile {
    static const Wy__TypeFD DefaultFd=WY__DEFAULT_FD;
    static const mode_t DefaultFType=0;
    static const Wy__CIF SysCif;

    // [Internal] Overwrite this->_m_fd to be exactly the given fd
    //            This member is dedicated for Wy__Std implement
    //
    //   --Warning-- 
    //   Function overwrites fd, make sure no resource is lost, and file type
    //   of fd is suitable for the class. And note the rule for ftype.
    //
    //  Note: This member is dedicated for WyByteFlow to initialize objects
    //        for standard devices. 
    //
    void _overwrite(Wy__TypeFD fd, mode_t ftype) WY__NOTHROW__;

    // [Cancel Point]
    //
    // [Syn] Open (::open) a file descriptor and overwite _m_fd/_m_ftype
    //
    // [Throw] Reply
    //         Wym_EFAULT      pathname pointed address is inaccessible
    //         Wym_EBADF       file not valid for the operation
    //         Wym_EACCES      Search permission is denied
    //         Wym_EINVAL      \fIf\fP invalid
    //         Wym_ENAMETOOLONG \fIpathname\fP too long
    //         Wym_ENOENT      \fIpathname\fP does not exist
    //         Wym_ENOTDIR     Entry in \fIpathname\fP is not a directory
    //         Wym_EMFILE      Too many open fd
    //         Wym_ENFILE      File table overflow
    //         Wym_ENOMEM      Not enough memory
    //         Wym_EEXIST      Pathname already exists
    //         Wym_EINTR       Interrupted by a signal
    //         Wym_ENOSPC      Device left no space for the operation
    //         Wym_ENXIO       No such device (device not ready)
    //         Wym_EROFS       Read-only file system
    //         Wym_ENODEV      No such device
    //         Wym_ELOOP       Too many symbolic links encountered
    //         ...             reply converted from ::open
    //
    // [Refer] ::open(const char*,int)
    //
    inline void _init(const char* pathname,int flags);

    #ifdef WY_DEBUG
    static int _obj_cnt;   // for tracking number of objects
    #endif

    Wy__TypeFD _m_fd;      // File descriptor

    // [Syn] Actual file type of _m_fd
    //
    //       0        default
    //       S_IFREG  regular file
    //       S_IFCHR  character device
    //       S_IFSOCK socket
    //       S_IFIFO  fifo
    //       S_IFDIR  directory file
    //       S_IFLNK  symbolic link
    //       S_IFBLK  block device
    //
    mode_t     _m_ftype;

    const Wy__CIF& _cif;

    // Hidden
    const WySysFile& operator=(const WySysFile& rhs);
    bool operator==(const WySysFile& rhs) const;
    bool operator!=(const WySysFile& rhs) const;
   
  protected:

    // note: These members are not public because implement does not know
    //       what the pathname is really meant (e.g. lnk).
    //
    WySysFile(const char* pathname,int f,const Wy__CIF& cif=SysCif);
    WySysFile(const WyStr& pathname,int f,const Wy__CIF& cif=SysCif);
    virtual WyRet reset(const char* pathname, int flags);
    virtual WyRet reset(const WyStr& pathname, int flags);

  public:
    static const char class_name[];
    WY_THROW_REPLY;

    // [Throw] Reply
    //         reserved
    //
    explicit  // no implicit conversion
    WySysFile(const Wy__CIF& cif=SysCif);

    // [Throw] Reply
    //         Wym_EBADF   \fIfh\fP not valid for the operation
    //         Wym_EMFILE  too many open devices
    //         Wym_ENOMEM  Not enough memory
    //
    // [Refer] ::dup(int)
    //
    WySysFile(WyFileHandle fh, const Wy__CIF& cif=SysCif);

    // [Throw] Reply
    //         error from ::dup(int)
    //
    // [Refer] ::dup(int)
    //
    WySysFile(const WySysFile& sysf, const Wy__CIF& cif=SysCif);

    // [Move Constructor]
    //
    // Note: sysf must be whole object except in initialization list
    //
    WySysFile(WySysFile& sysf,Wy::ByMove_t) WY__TSPC();

    // Note: Not Cancellable
    // 
    virtual ~WySysFile();

    bool is_default(void) const WY__TSPC();
    WyFileHandle fh(void) const WY__TSPC();
    mode_t ftype(void) const WY__TSPC();
    WyRet stat(WyFileStat& filestat) const;

    // [Cancel Point]
    //
    virtual WyRet reset(void);
    virtual WyRet reset(WyFileHandle fh);
    virtual WyRet reset(const WySysFile& sysf);

    virtual WyRet _swap(WySysFile& sysf) WY__TSPC();

    virtual WySysFile* _alloc(WyRet&) const;

    // [Cancel Point]
    //
    // [Syn] Close the file descriptor
    //
    // [Ret] 0     OK
    //       EBADF fd is not a valid open fd
    //       EINTR interrupted by a signal (fd is closed)
    //       EIO   I/O error
    //
    // [Refer] ::close(int)
    //
    static int wy_close(const Wy__TypeFD& fd);

    Wy__TypeFD wy_fd(void) const WY__NOTHROW__
         { return _m_fd; };

    #ifdef WY_DEBUG
    static int wy_obj_cnt(void) WY__TSPC()
         { return _obj_cnt; };
    #endif

    friend class Wy__Std;    // Wy__Std accesses _cif directly, and _overwrite

    // The chk file will access protected ctor/reset(..)..
    friend void ck_sysfile_1(void);
    friend void ck_byteflow_1(void);
};

#endif
