/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003
*/

#define WYLIB_SOURCE
#include "wytimespec.h"
#include "wy_atdestroy.h"
#include "wymutex.h"
#include "wyonce.h"
#include <limits>
#include <ctime>
#include <cmath>              // for modf
#include <cstdlib>
#include <new>

//#undef _POSIX_TIMERS        // for check

#ifndef _POSIX_TIMERS
#include <sys/time.h>

// Mutex dedicated for sycronizing system time access (gettimeofday)
//
static WyMutex wy__cmtx;
#endif

// [Syn] Test overflow 
//
// [Ret] true: a+b overflows
//      false: otherwise
//
template<typename T>
static bool wy__add_ovf(T a, T b)
{
 if(std::numeric_limits<T>::is_integer) {
   return (b>=0)? (a>std::numeric_limits<T>::max()-b)
                : (a<std::numeric_limits<T>::min()-b);
 } else {
   // not expected to reach here
   WY_TERMINATE("");
 }
};

// [Syn] Test overflow 
//
// [Ret] true: a-b overflows
//      false: otherwise
//
template<typename T>
static bool wy__sub_ovf(T a, T b)
{
 if(std::numeric_limits<T>::is_integer) {
   return (b>=0)? (a<std::numeric_limits<T>::min()+b)    // min-0
                : (a>std::numeric_limits<T>::max()+b);   // 0-min
 } else {
   // not expected to reach here
   WY_TERMINATE("");
 }
};

static const char WY__EToken('e');  // search for WY__EToken
                                    // many are in string literal

const char WyTimeSpec::class_name[]="WyTimeSpec";

class WyTimeSpec::_Assertion {
  public:
    _Assertion() {

        // Wy_Second is time_t
        //
        if((std::numeric_limits<Wy_Second>::max()!=
            std::numeric_limits<time_t>::max())||
           (std::numeric_limits<Wy_Second>::min()!=
            std::numeric_limits<time_t>::min())) {
          WY_TERMINATE("");
        }

        // Wy_Second is representable by long
        //
        if((std::numeric_limits<Wy_Second>::max()>
            std::numeric_limits<long>::max())||
           (std::numeric_limits<Wy_Second>::min()<
            std::numeric_limits<long>::min())) {
          WY_TERMINATE("");
        }

        // Wy_Nano is long
        //
        if((std::numeric_limits<Wy_Nano>::max()!=
            std::numeric_limits<signed long>::max())||
           (std::numeric_limits<Wy_Nano>::max()!=
            std::numeric_limits<signed long>::max())) {
          WY_TERMINATE("");
        }

        // Default value of WyTimeSpec is 0.0
        //
        WyTimeSpec tmp;
        if((tmp.second()!=0)||(tmp.nano()!=0)) {
          WY_TERMINATE("");
        }
    };
} static const wy__do_not_use_me__;

WyRet WyTimeSpec::_normalize(long& sec, long& nsec) WY__TSPC()
{
 const std::ldiv_t fr=std::ldiv(nsec,_Giga);
 const Wy_Second ts=sec+fr.quot;

 if(wy__add_ovf(sec,fr.quot)) {
   WY_RETURN(Wym_ERANGE);   // result not representable
 }

 // make sec/nano the same sign
 if(ts>0) {
   if(fr.rem<0) {
     sec=ts-1;
     nsec=fr.rem+_Giga;
     return(OK);
   }
   // FALL_THROUGH
 } else if(ts<0) {
   if(fr.rem>0) {
     sec=ts+1;
     nsec=fr.rem-_Giga;
     return(OK);
   }
   // FALL_THROUGH
 }
 sec=ts;
 nsec=fr.rem;
 return(OK);
};

WyRet WyTimeSpec::_normalize(long long int& sec, long long int& nsec) WY__TSPC()
{
 const std::lldiv_t fr=std::lldiv(nsec,_Giga);
 const long long ts=sec+fr.quot;

 if(wy__add_ovf(sec,fr.quot)) {
   WY_RETURN(Wym_ERANGE);   // result not representable
 }

 // make sec/nano the same sign
 if(ts>0) {
   if(fr.rem<0) {
     sec=ts-1;
     nsec=fr.rem+_Giga;
     return(OK);
   }
   // FALL_THROUGH
 } else if(ts<0) {
   if(fr.rem>0) {
     sec=ts+1;
     nsec=fr.rem-_Giga;
     return(OK);
   }
   // FALL_THROUGH
 }
 sec=ts;
 nsec=fr.rem;
 return(OK);
};

// [Internal] Initialize object to contain the text indicated number
//            The text is blen characters pointed by buf in the form:
//            [+-]ddd.ddd[(eE)[+-]ddd], radix is 10
//
//            second()= converted from the text
//            nano()  = ditto
// [Ret] OK
//       Wym_EFAULT
//       Wym_EBADMSG
//       Wym_ERANGE
//
inline WyRet WyTimeSpec::_init(const char* buf,size_t blen) WY__TSPC()
{
 WyTimeSpec tm;
 const WyRet r( Wy::_strnum(tm,NULL,WyCSeg(buf,blen)) );
 if(r!=OK) {
   WY_RETURN(r);
 }
 _tspc=tm._tspc;
 return(OK);
};

WyTimeSpec::WyTimeSpec(Wy_Second sec, Wy_Nano nsec) WY__TSPC(Reply)
{
 const WyRet r(_normalize(sec, nsec)); 
 if(r!=OK) {
   WY_THROW( Reply(r) );
 }
 _tspc.tv_sec=sec;
 _tspc.tv_nsec=nsec;
};

WyTimeSpec::WyTimeSpec(const WyCSeg& cseg) WY__TSPC(Reply)
{
 const WyRet r( this->_init(cseg.begin(),cseg.size()) );
 if(r!=OK) {
   WY_THROW( Reply(r) );
 }
};

WyTimeSpec::WyTimeSpec(double fvalue) WY__TSPC(Reply)
{
 double it,fr=std::modf(fvalue,&it);
 if((it> static_cast<double>( std::numeric_limits<Wy_Second>::max() ))||
    (it< static_cast<double>( std::numeric_limits<Wy_Second>::min() ))) {
   WY_THROW( Reply(Wym_ERANGE) );
 }
 _tspc.tv_sec =static_cast<Wy_Second>(it);
 _tspc.tv_nsec=static_cast<Wy_Nano>(fr*WY_CONST_GIGA);
};

WyRet WyTimeSpec::reset(Wy_Second sec, Wy_Nano nano) WY__TSPC()
{
 const WyRet r( _normalize(sec, nano) ); 
 if(r!=OK) {
   WY_RETURN(r);
 }
 _tspc.tv_sec=sec;
 _tspc.tv_nsec=nano;
 return(OK);
};

WyRet WyTimeSpec::reset(const WyCSeg& cseg) WY__TSPC()
{
  WY_RETURN( this->_init(cseg.begin(),cseg.size()) );
};

/*
void WyTimeSpec::reset(double fvalue) WY__TSPC()
{ 
  this->_init(fvalue);
};

void WyTimeSpec::reset(float fvalue) WY__TSPC()
{ 
  this->_init(fvalue);
};
*/

WyTimeSpec& WyTimeSpec::operator =(const WyCSeg& cseg) WY__TSPC(Reply)
{
  const WyRet r( this->_init(cseg.begin(),cseg.size()) );
  if(r!=OK) {
    WY_THROW( Reply(r) );
  }
  return(*this);
};

const WyTimeSpec WyTimeSpec::operator -() const WY__TSPC(Reply)
{
 if(wy__sub_ovf(std::time_t(0),this->_tspc.tv_sec)) {
   WY_THROW( Reply(Wym_EMATHNEG) );
 }
 return WyTimeSpec(-_tspc.tv_sec,-_tspc.tv_nsec);
};

WyTimeSpec & WyTimeSpec::operator +=(const WyTimeSpec &rhs) WY__TSPC(Reply)
{
 if(wy__add_ovf(this->_tspc.tv_sec,rhs._tspc.tv_sec)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec+=rhs._tspc.tv_sec;
 v._tspc.tv_nsec+=rhs._tspc.tv_nsec;

 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=OK) {
   WY_THROW( Reply(r) );
 }
 _tspc=v._tspc;
 return(*this);
};

WyTimeSpec & WyTimeSpec::operator -=(const WyTimeSpec &rhs) WY__TSPC(Reply)
{
 if(wy__sub_ovf(this->_tspc.tv_sec,rhs._tspc.tv_sec)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec-=rhs._tspc.tv_sec;
 v._tspc.tv_nsec-=rhs._tspc.tv_nsec;

 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=OK) {
   WY_THROW( Reply(r) );
 }
 _tspc=v._tspc;
 return(*this);
};

const WyTimeSpec WyTimeSpec::operator +(const WyTimeSpec &rhs) const WY__TSPC(Reply)
{
 if(wy__add_ovf(this->_tspc.tv_sec,rhs._tspc.tv_sec)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec+=rhs._tspc.tv_sec;
 v._tspc.tv_nsec+=rhs._tspc.tv_nsec;

 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=OK) {
   WY_THROW( Reply(r) );
 }
 return(v);
};

const WyTimeSpec WyTimeSpec::operator -(const WyTimeSpec &rhs) const WY__TSPC(Reply)
{
 if(wy__sub_ovf(this->_tspc.tv_sec,rhs._tspc.tv_sec)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec-=rhs._tspc.tv_sec;
 v._tspc.tv_nsec-=rhs._tspc.tv_nsec;

 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=OK) {
   WY_THROW( Reply(r) );
 }
 return(v);
};

bool WyTimeSpec::operator >(const WyTimeSpec &rhs) const WY__TSPC()
{
 if(_tspc.tv_sec>rhs._tspc.tv_sec) {
   return true;
 } else if(_tspc.tv_sec<rhs._tspc.tv_sec) {
   return false;
 } else {}

 if(_tspc.tv_nsec>rhs._tspc.tv_nsec) {
   return true;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator >=(const WyTimeSpec &rhs) const WY__TSPC()
{
 if(_tspc.tv_sec>rhs._tspc.tv_sec) {
   return true;
 } else if(_tspc.tv_sec<rhs._tspc.tv_sec) {
   return false;
 } else {}

 if(_tspc.tv_nsec>=rhs._tspc.tv_nsec) {
   return true;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator <(const WyTimeSpec &rhs) const WY__TSPC()
{
 if(_tspc.tv_sec<rhs._tspc.tv_sec) {
   return true;
 } else if(_tspc.tv_sec>rhs._tspc.tv_sec) {
   return false;
 } else {}

 if(_tspc.tv_nsec<rhs._tspc.tv_nsec) {
   return true;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator <=(const WyTimeSpec &rhs) const WY__TSPC()
{
 if(_tspc.tv_sec<rhs._tspc.tv_sec) {
   return true;
 } else if(_tspc.tv_sec>rhs._tspc.tv_sec) {
   return false;
 } else {}

 if(_tspc.tv_nsec<=rhs._tspc.tv_nsec) {
   return true;
 } else {
   return false;
 }
};

const WyTimeSpec WyTimeSpec::abs(void) const WY__TSPC(Reply)
{
 WyTimeSpec v; 
 if(_tspc.tv_sec>0) {
   v._tspc.tv_sec=_tspc.tv_sec;
   v._tspc.tv_nsec=_tspc.tv_nsec;
 } else if(_tspc.tv_sec<0) {
   if(wy__sub_ovf(std::time_t(0),_tspc.tv_sec)) {
     WY_THROW( Reply(Wym_EMATHNEG) );
   }
   v._tspc.tv_sec=-_tspc.tv_sec;
   v._tspc.tv_nsec=-_tspc.tv_nsec;
 } else {
   // v.tv_sec=0;  already 0
   if(_tspc.tv_nsec>=0) {
     v._tspc.tv_nsec=_tspc.tv_nsec;
   } else {
     v._tspc.tv_nsec=-_tspc.tv_nsec;
   }
 }
 return(v);
};

WyRet WyTimeSpec::add(Wy_Second sec, Wy_Nano nsec) WY__TSPC()
{
 WyRet r( _normalize(sec,nsec) );
 if(r!=OK) {
   WY_RETURN(r);
 }
 if(wy__add_ovf(_tspc.tv_sec,sec)) {
   WY_RETURN(Wym_ERANGE);
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec+=sec;
 v._tspc.tv_nsec+=nsec;

 if((r=_normalize(v._tspc.tv_sec,v._tspc.tv_nsec))!=OK) {
   WY_RETURN(r);
 }
 _tspc=v._tspc;
 return(OK);
};

WyRet WyTimeSpec::sub(Wy_Second sec, Wy_Nano nsec) WY__TSPC()
{ 
 WyRet r( _normalize(sec,nsec) );
 if(r!=OK) {
   WY_RETURN(r);
 }
 if(wy__sub_ovf(_tspc.tv_sec,sec)) {
   WY_RETURN(Wym_ERANGE);
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec-=sec;
 v._tspc.tv_nsec-=nsec;

 if((r=_normalize(v._tspc.tv_sec,v._tspc.tv_nsec))!=OK) {
   WY_RETURN(r);
 }
 _tspc=v._tspc;
 return(OK);
};

WyRet WyTimeSpec::mul(long int n) WY__TSPC()
{
 long long ll((long long)_tspc.tv_sec*n);
 
 if( (ll>(long long)std::numeric_limits<Wy_Second>::max())
   ||(ll<(long long)std::numeric_limits<Wy_Second>::min())) {
   WY_RETURN( Wym_ERANGE );
 }
 std::lldiv_t dt=std::lldiv((long long)_tspc.tv_nsec*n,_Giga);
 ll+=dt.quot;
 if( (ll>(long long)std::numeric_limits<Wy_Second>::max())
   ||(ll<(long long)std::numeric_limits<Wy_Second>::min())) {
   WY_RETURN( Wym_ERANGE );
 }
 
 _tspc.tv_sec=static_cast<Wy_Second>(ll);
 _tspc.tv_nsec=static_cast<Wy_Nano>(dt.rem);
 return(OK);
};

WyRet WyTimeSpec::mul(const WyTimeSpec& n) WY__TSPC()
{
 long long llnano((long long)_tspc.tv_nsec*n._tspc.tv_nsec);
 std::lldiv_t dt=std::lldiv((long long)llnano,(long long)WY_CONST_GIGA);
 llnano=(dt.quot);

 long long tmp((long long)_tspc.tv_sec*n._tspc.tv_nsec);
 dt=std::lldiv(tmp,(long long)WY_CONST_GIGA);
 long long llsec(dt.quot);
 llnano+=dt.rem;

 tmp=(long long)_tspc.tv_nsec*n._tspc.tv_sec;
 dt=std::lldiv(tmp,(long long)WY_CONST_GIGA);
 llsec+=dt.quot;
 llnano+=dt.rem;

 dt=std::lldiv(llnano,(long long)WY_CONST_GIGA);
 llnano=dt.rem;
 llsec+=dt.quot;

 tmp=(long long)_tspc.tv_sec*n._tspc.tv_sec;

 if(wy__add_ovf(tmp,llsec)) {
   WY_RETURN(Wym_ERANGE);   // result not representable
 }
 llsec+=tmp;
 if( (tmp>(long long)std::numeric_limits<Wy_Second>::max())
   ||(tmp<(long long)std::numeric_limits<Wy_Second>::min())) {
   WY_RETURN( Wym_ERANGE );
 }
 _tspc.tv_sec = llsec;
 _tspc.tv_nsec= llnano;
 return(OK);
};

WyRet WyTimeSpec::div(long n) WY__TSPC()
{
 if(n==0) {
   WY_RETURN( Wym_EDIVZERO );
 }
 const std::ldiv_t dt=std::ldiv(_tspc.tv_sec,n);

 _tspc.tv_sec=dt.quot;
 _tspc.tv_nsec= (((long long)dt.rem*(long long)_Giga)+_tspc.tv_nsec)/n;
 return(OK);
};
/*
#include <iostream>
WyRet WyTimeSpec::div(const WyTimeSpec& n) WY__TSPC()
{
 if((n._tspc.tv_sec==0)&&(n._tspc.tv_nsec==0)) {
   WY_RETURN( Wym_EDIVZERO );
 }
 const long long Dvr((long long)n._tspc.tv_sec*(long long)_Giga+n._tspc.tv_nsec);
 std::lldiv_t dt=std::lldiv(
              (long long)_tspc.tv_sec*(long long)_Giga +_tspc.tv_nsec, Dvr);

 long long wn(dt.quot);     // whole number
 //std::cerr << wn << '/' << Dvr << "= " << dt.quot << ", " << dt.rem << '\n';
 dt=std::lldiv(dt.rem,Dvr);  // r -> quot[Dvr]+rem   (rem<Dvr)

 long long t2(dt.quot*_Giga);
 long long t3(dt.rem*_Giga);

 std::lldiv_t dt2=std::lldiv((long long)_Giga,Dvr);

 _tspc.tv_sec = wn;
 _tspc.tv_nsec= dt.quot;
 return(OK);
};*/

//-------------------------------------
WyStr Wy::wrd(const WyTimeSpec& tm)
{
 WyStr str;
 WyRet r( Wy::_mkstr(str,tm,10,WyTimeSpec::_GigaDigit10) );
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 // remove trailing zero
 for(int i=(int)str.size()-1; i>=0; --i) {
   if(str[i]!='0') {
     if(str[i]=='.') {
       i+=2;
     } else {
       ++i;
     }
     if((r=str.resize(i))!=OK) {
       WY_THROW( WyStr::Reply(r) );
     }
     break;
   }
 }
 return(str);
};

WyStr Wy::wrd(const WyTimeSpec& num, int radix, size_t frdig)
{
 WyStr str;
 const WyRet r( Wy::_mkstr(str,num,radix,frdig) );
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyTimeSpec Wy::now(void)
try {
#ifdef _POSIX_TIMERS
 struct timespec tm_now;
 if(::clock_gettime(CLOCK_REALTIME,&tm_now)!=0) {
   // man. only EINVAL which should not happen here.
   WY_THROW( WyRet(WyReply(errno)) );
 }

 //
 // note: WyTimeSpec ctor is assumed not to throw in this condition
 //
 return WyTimeSpec(tm_now.tv_sec,tm_now.tv_nsec);
#else
 struct ::timeval tm_now;
 WyRet r;
 if((r=wy__cmtx.lock())!=OK) {
   WY_THROW(r);
 }
 //
 // guess this function is not thread-safe, so a mutex supplied.
 //
 if(::gettimeofday(&tm_now,0)!=0) {
   //
   // man indicates all errors are associated with time zone, which is not 
   // used int time(), so it is not supposed to happen here.
   //
   const int v=errno;
   wy__cmtx.unlock();    // not check the error, for result is the same.
   WY_THROW( WyRet(WyReply(v)) );
 }
 if((r=wy__cmtx.unlock())!=OK) {
   WY_THROW(r);
 }

 //
 // note: WyTimeSpec ctor is assumed not to throw in this condition
 //
 return WyTimeSpec(tm_now.tv_sec,tm_now.tv_usec*1000);
#endif
}
catch(const WyTimeSpec::Reply& e) {
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
 WY_THROW( WyRet(e) );
};

WyRet Wy::set_systime(const WyTimeSpec& tm)
{
#ifdef _POSIX_TIMERS
 if(::clock_settime(CLOCK_REALTIME,&tm.wy_timespec())!=0) {
   const int v=errno;
   switch(v) {
     case EPERM: WY_RETURN(Wym_EPERM);
     case EINVAL: WY_RETURN(Wym_ERANGE); // guess this reply
     case EFAULT: WY_THROW(WyRet(Wym_EFAULT)); // should not happen
     default:    WY_RETURN( WyRet(WyReply(v)) );  
   };
 }
 return(OK);
#else
 WyRet r;
 struct timeval tm_now;

 if((r=wy__cmtx.lock())!=OK) {
   WY_THROW(r);
 }
 tm_now.tv_sec=wytm.second();
 tm_now.tv_usec=static_cast<long>(wytm.nano()/1000);

 // guess this function is not thread-safe, so a mutex supplied
 const int en=(::settimeofday(&tm_now,0)==0)? 0:errno;
 if((r=wy__cmtx.unlock())!=OK) {
   WY_THROW(r);
 }
 switch(en) {
   case 0: return(OK);
   case EPERM: WY_RETURN(Wym_EPERM);
   case EINVAL: // FALL_THROUGH
   case EFAULT: WY_THROW(WyRet(Wym_EFAULT));  // should not happen
   default:    WY_RETURN( WyRet(WyReply(en)) ); 
 };
#endif
};

//----------------------------------------------------------------------------
//
WyRet Wy::sleep(WyTimeSpec tsp)
{
 WY__CANCEL_POINT;
 const int v( ::nanosleep(&tsp.wy_timespec(),NULL) );
 WY__CANCEL_POINT;
 if(v!=0) {
   WY_RETURN(errno);
 }
 return(OK);
};

WyRet Wy::sleep(WyTimeSpec req, WyTimeSpec& rem)
{
 WY__CANCEL_POINT;
 const int v( ::nanosleep(&req.wy_timespec(),&rem.wy_timespec()) );
 WY__CANCEL_POINT;
 if(v!=0) {
   WY_RETURN(errno);
 }
 return(OK);
};

// [Interna] Dedicated for sleep_till
//           Unlock mtx and destroy mtx/cond
//
void static wy__sleep_till_cleanup(::pthread_cond_t& cond, ::pthread_mutex_t& mtx)
{
 ::pthread_mutex_unlock(&mtx);
 ::pthread_mutex_destroy(&mtx);
 ::pthread_cond_destroy(&cond);
};

WyRet Wy::sleep_till(WyTimeSpec wtm)
{
 ::pthread_cond_t cond;
 ::pthread_mutex_t mtx;

 ::pthread_mutex_init(&mtx,NULL);   // manual: always ok
 const int v=::pthread_mutex_lock(&mtx);
 if(v!=0) {
   ::pthread_mutex_destroy(&mtx);
   WY_THROW( WyRet(WyReply(v)) );  // internal error
 }
 ::pthread_cond_init(&cond,NULL);   // manual: always ok
 Wy_AtDestroy2<void,::pthread_cond_t&,::pthread_mutex_t&> 
   rrid(wy__sleep_till_cleanup,cond,mtx);

 WY_RETURN( WyReply(::pthread_cond_timedwait(&cond,&mtx,&wtm.wy_timespec())));
};

// [Internal] Make str in float notation [-]ddd.ddd
static WyRet wy___ms_float(WyStr &str,const WyTimeSpec& num,int radix,size_t frdig)
{
 const Wy_Second tm_secs=num.second();
 const Wy_Nano tm_nano=num.nano();

 WyRet r;
 WyStr str_1;

 if(tm_nano==0) {
   if((r=Wy::_catstr(str_1,tm_secs,radix))!=OK) {
     WY_RETURN(r);
   }
   if(frdig==0) {
     WY_RETURN( str.reset(str_1) );
     return(OK);
   }
   if((r=str_1.append('.'))!=OK) {
     WY_RETURN(r);
   }
   if((r=str_1.append(frdig,'0'))!=OK) {
     WY_RETURN(r);
   }
   WY_RETURN( str.reset(str_1) );
 }

 WyStr str_2,str_tmp;

 // Make str_1 the representation of tv_sec
 //      str_2 the representation of tv_nsec/GiGa
 //
 if(tm_nano<0) {
   if(tm_secs<0) {
     if((r=Wy::_catstr(str_1,tm_secs,radix))!=OK) {
       WY_RETURN(r);
     }
   } else {
     if(tm_secs!=0) {
       // assertion(tv_sec==0) failed
       WY_THROW( WyRet(Wym_ERANGE) );
     }
     str_1.reset("-0");  // min_capacity>2
   }
   if((r=Wy::_catstr(str_tmp,-tm_nano,radix))!=OK) {
     WY_RETURN(r);
   }
 } else {
   if((r=Wy::_catstr(str_1,tm_secs,radix))!=OK) {
     WY_RETURN(r);
   }
   if((r=Wy::_catstr(str_tmp,tm_nano,radix))!=OK) {
     WY_RETURN(r);
   }
 }

 if(frdig>0) {
   // assert( str_tmp[0]!=0 )
   const int lz=WyTimeSpec::_GigaDigit10-(int)str_tmp.size();
   if(lz<0) {
     // assertion failure: 
     // str_tmp.size() should be 1-9
     WY_THROW( WyRet(Wym_ERANGE) );
   }
   if((r=str_2.append('.'))!=OK) {
     WY_RETURN(r);
   }
   if((r=str_2.append(lz,'0'))!=OK) {  // prepend leading zero's
     WY_RETURN(r);
   }
   if((r=str_2.append(str_tmp))!=OK) {
     WY_RETURN(r);
   }
   if((r=str_2.resize(frdig+1,'0'))!=OK) { // one '.' already in str_2
     WY_RETURN(r);
   }
   if((str_1.append(str_2))!=OK) {
     WY_RETURN(r);
   }
 }
 WY_RETURN( str.reset(str_1) );
};

WyRet Wy::_mkstr(WyStr &str,const WyTimeSpec& num,int radix,size_t frdig)
try {
 if(radix==10) {
   WY_RETURN(wy___ms_float(str,num,radix,frdig));
 } else if(radix!=0) {
   WY_RETURN(Wym_EINVAL);
 } else {
   // FALLTHROGH (radix==0)
 }

 const int Radix=10;                 // float type always uses radix 10
 const Wy_Second tm_secs=num.second();
 const Wy_Nano tm_nano=num.nano();
 WyRet r;

 if(tm_secs==0) {
   WyStr str1,str2;
   if(tm_nano==0) {
     if(frdig==0) {
       WY_RETURN( str.reset("0e+0") );   // WY__EToken
     }
     str1.append("0.");
     str1.append(frdig,'0');
     str1+="e+0";   // WY__EToken
     WY_RETURN(str.reset(str1));
   }
   if(tm_nano<0) {
     str1="-";
     if((r=_mkstr(str2,-tm_nano,Radix))!=OK) {
       WY_RETURN(r);
     }
   } else {
     if((r=_mkstr(str2,tm_nano,Radix))!=OK) {
       WY_RETURN(r);
     }
   }
   int exp=str2.size()-WyTimeSpec::_GigaDigit10-1;
   if(exp>=0) {
     // ttv_nsec has more than 9 digits
     WY_THROW( WyRet(Wym_ERANGE) );
   }
   str1+=str2[0];
   if(frdig>0) {
     if((r=str2.resize(frdig+1,'0'))!=OK) {
       WY_RETURN(r);
     }
     str1+='.';
     if((r=str1.append(str2.cseg(1,str2.size())))!=OK) {
       WY_RETURN(r);
     }
   }

   str1+='e';         // WY__EToken (strange? compiler complains)
   if((r=_mkstr(str2,exp,Radix))!=OK) {
     WY_RETURN(r);
   }
   str1+=str2;
   WY_RETURN(str.reset(str1));
 } else {
   WyStr str1,str2;
   if((r=_mkstr(str1,tm_secs,Radix))!=OK) {
     WY_RETURN(r);
   }
   // insert '.' to str1 right after the first digit
   size_t dig0_idx;
   for(dig0_idx=0; dig0_idx<str1.size(); ++dig0_idx) {
     const char dch=str1[dig0_idx];
     if((dch>='0')&&(dch<='9')) {
       break;
     }
   }
   if(dig0_idx>=str1.size()) {
     WY_THROW( WyRet(Wym_ERANGE) );
   }
   size_t exp=str1.size()-dig0_idx-1;
   if(frdig>0) {
     if((r=str1.insert(dig0_idx+1,1,'.'))!=OK) {
       WY_RETURN(r);
     }
     // make the fraction
     if((r=_mkstr(str2,(tm_nano<0)?-tm_nano:tm_nano,Radix))!=OK) {
       WY_RETURN(r);
     }
     {
       const int apz=WyTimeSpec::_GigaDigit10-(int)str2.size();
       if(apz<0) {
         WY_THROW( WyRet(Wym_ERANGE) );
       }
       if((r=str2.insert(0,apz,'0'))!=OK) {
         WY_RETURN(r);
       }
     }
     str1+=str2;
     dig0_idx+=frdig+2;   // total length str1 should be
     if((r=str1.resize(dig0_idx,'0'))!=OK) {
       WY_RETURN(r);
     }
   }

   str2.reset("e+");   // WY__EToken
   if((r=_catstr(str2,exp,Radix))!=OK) {
     WY_RETURN(r);
   }
   str1+=str2;
   WY_RETURN(str.reset(str1));
 }
}
catch(const WyStr::Reply& e) {
  return(e);
}
catch(const std::bad_alloc&) {
  WY_RETURN(Wym_ENOMEM); 
};

// [Internal] Provide common functions for _strnum(WyTimeSpec&,..)
//
class Wy__P10 {
  public:
     static const size_t MaxP10=18;
  private:
    typedef signed long long value_type;   // better able to hold all precision
    static const int GiDig10=WyTimeSpec::_GigaDigit10;
    static const value_type Giga=WY_CONST_GIGA;    // 1x10^9
    static const value_type p10tab[MaxP10+1];
  public:
    // [Syn] Get power of 10
    // 
    //  Note: use MaxP10 for limit of exp
    // 
    // [Ret] 10^exp
    //
    inline static value_type p10(int exp)
         { return p10tab[exp]; };

    // [Syn] Convert input number (taken as num/10^exp10) to
    //       another representation, as follows:
    //
    //       Let input_number= num/10^exp10, then upon returned,
    //       res1= input_number/Giga
    //       res2= input_number%Giga
    //
    //       Note: result value less than 1/Giga is ignored
    //
    // [Ret] OK
    //       Wym_EBADMSG  exp10>MaxP10+GiDig10 (for cooperate with _strnum)
    //       Wym_ERANGE   overflow occurred in operation 
    //
    static WyRet norm_nano(value_type& res1, value_type& res2,
                           value_type num, int exp10)
         {
           if((std::numeric_limits<value_type>::max()!=
                          std::numeric_limits<long long>::max())||
              (std::numeric_limits<value_type>::min()!=
                          std::numeric_limits<long long>::min())) {
              WY_TERMINATE("");  // as lldiv_t is used, check for signed long long.
           }
           if(exp10<0) {
             if(exp10>=-GiDig10) {
               std::lldiv_t dt;
               WyRet r;
               value_type tmp;
               const size_t ddp=GiDig10+exp10;
               if(ddp>Wy__P10::MaxP10) {
                 // assertion failed (GiDig10<Wy__P10::MaxP10)
                 WY_THROW(WyRet(Wym_ERANGE)); 
               }
               if((r=Wy::wy_mul(tmp,num,p10tab[ddp]))!=OK) {
                 WY_RETURN(r);
               }
               dt=std::lldiv(tmp,Giga);
               res1=dt.quot;
               res2=dt.rem;
             } else if(exp10>=-(int)MaxP10) {
               const size_t ddp=-exp10-GiDig10;
               std::lldiv_t dt;
               if(ddp>MaxP10) {
                 WY_RETURN(Wym_EBADMSG);
               }
               dt=std::lldiv(num,p10tab[ddp]);
               res1=0;
               res2=dt.quot;  // dt.rem ignored
             } else {
               res1=res2=0;
             }
           } else {
             WyRet r;
             value_type tmp;
             if(exp10>(int)MaxP10) {
               WY_RETURN(Wym_EBADMSG);
             }
             if((r=Wy::wy_mul(tmp,num,p10tab[exp10]))!=OK) {
               WY_RETURN(r);
             }
             res1=tmp;
             res2=0;
           }
           return(OK);
         };
};

const Wy__P10::value_type Wy__P10::p10tab[MaxP10+1]=
{ 1LL,10LL,100LL,1000LL,10000LL,100000LL,1000000LL, 10000000LL, // 10^0 - 10^7
  100000000LL,          // 10^8
  1000000000LL,         // 10^9
  10000000000LL,        // 10^10
  100000000000LL,       // 10^11
  1000000000000LL,      // 10^12
  10000000000000LL,     // 10^13
  100000000000000LL,    // 10^14
  1000000000000000LL,   // 10^15
  10000000000000000LL,  // 10^16
  100000000000000000LL, // 10^17
  1000000000000000000LL // 10^18
};

// S::= S1|S2|S3
// S1::= [+-]ddd[<Exp>]
// S2::= [+-].ddd[<Exp>]
// S3::= [+-]ddd.[ddd][<Exp>]
// Exp::= [Ee][+-]ddd
//
// Wym_EBADMSG   failed processing string representation
//
WyRet Wy::_strnum(WyTimeSpec& res, const char** endptr, WyCSeg cseg)
{
 typedef long long TmpInt;   // better able to hold all precision
 const int Radix=10;
 TmpInt ll_sec;
 TmpInt ll_nsec;
 int exptv_nsec;  
 const char* cptr;
 size_t clft;

 WyRet r( Wy::wy_strnum2(ll_sec,ll_nsec,exptv_nsec,&cptr,cseg.begin(),cseg.size()) );

 // SRC_NUM= ll_sec + ll_nsec*10^exptv_nsec (exptv_nsec<=0)
 //
 if(r==OK) {
   if(endptr!=0) {
     *endptr=cptr;
   }
   if((ll_sec>(TmpInt)std::numeric_limits<Wy_Second>::max())||
      (ll_sec<(TmpInt)std::numeric_limits<Wy_Second>::min())) {
     WY_RETURN(Wym_ERANGE);
   }
   // normalize ll_nsec
   TmpInt tmp;
   if((r=Wy__P10::norm_nano(tmp, ll_nsec,ll_nsec,exptv_nsec))!=OK) {
     WY_RETURN(r);
   }
   if(tmp!=0) {
     // assertion failed
     WY_THROW(WyRet(Wym_ERANGE));
   }
   if((ll_nsec>=TmpInt(WY_CONST_GIGA))||
      (ll_nsec<=-TmpInt(WY_CONST_GIGA))) {
     // assertion failed
     WY_THROW(WyRet(Wym_ERANGE));
   }
   // S= [+-]ddd[.ddd]
   //  = [+-]ddd.
   //  = [+-].ddd
   WY_RETURN( res.reset(Wy_Second(ll_sec),Wy_Nano(ll_nsec)) );
 }
 if(r!=Wym_EBADMSG) {
   if(endptr!=0) {
     *endptr=cptr;
   }
   WY_RETURN(r);
 }
 clft=cseg.size()-(cptr-cseg.begin());
 if(clft<2) {
   if(endptr!=0) {
     *endptr=cptr;
   }
   WY_RETURN(Wym_EBADMSG);
 }
 if((*cptr!='e')&&(*cptr!='E')) {
   if(endptr!=0) {
     *endptr=cptr;
   }
   WY_RETURN(Wym_EBADMSG);
 }
 --clft;
 ++cptr;
 int tt_exp;
 r=Wy::_strnum(tt_exp,&cptr,WyCSeg(cptr,clft),Radix);
 if(endptr!=0) {
   *endptr=cptr;
 }
 if(r!=OK) {
   if(r==Wym_ENOENT) {
     WY_RETURN(Wym_EBADMSG);
   }
   WY_RETURN(r);
 }
 
 // SRC_NUM= (ll_sec + ll_nsec*10^exptv_nsec)*10^tt_exp   (exptv_nsec<=0)
 //
 if(tt_exp>=0) {
   if(tt_exp>(int)Wy__P10::MaxP10) {
     WY_RETURN(Wym_ERANGE);
   }
   TmpInt tmp;
   if((r=Wy::wy_mul(ll_sec,ll_sec,Wy__P10::p10(tt_exp)))!=OK) {
     WY_RETURN(r);
   }
   if((r=Wy__P10::norm_nano(tmp,ll_nsec,ll_nsec,exptv_nsec+tt_exp))!=OK) {
     WY_RETURN(r);
   }
   if((r=Wy::wy_add(ll_sec,ll_sec,tmp))!=OK) {
     WY_RETURN(r);
   }
   if((ll_sec>(TmpInt)std::numeric_limits<Wy_Second>::max())||
      (ll_sec<(TmpInt)std::numeric_limits<Wy_Second>::min())) {
     WY_RETURN(Wym_ERANGE);
   }
   WY_RETURN( res.reset(Wy_Second(ll_sec),Wy_Nano(ll_nsec)) );
 } else {
   const int tt_exp_abs=-tt_exp;
   TmpInt tmp;
   if(tt_exp_abs>(int)Wy__P10::MaxP10) {
     WY_RETURN(Wym_ERANGE);
   }
   std::lldiv_t dt=std::lldiv(ll_sec,Wy__P10::p10(tt_exp_abs));
   ll_sec=dt.quot;

   {
     const int dpp=-exptv_nsec;
     if(dpp>(int)Wy__P10::MaxP10) {
       WY_RETURN(Wym_EBADMSG);
     }
     if((r=Wy::wy_mul(tmp,dt.rem,Wy__P10::p10(dpp)))!=OK) {
       WY_RETURN(r);
     }
   }
   if((r=Wy::wy_add(tmp,tmp,ll_nsec))!=OK) {
     WY_RETURN(r);
   }

   if((r=Wy__P10::norm_nano(tmp,ll_nsec,tmp,tt_exp+exptv_nsec))!=OK) {
     WY_RETURN(r);
   }
   if((r=Wy::wy_add(ll_sec,ll_sec,tmp))!=OK) {
     WY_RETURN(r);
   }

   if((ll_sec>(TmpInt)std::numeric_limits<Wy_Second>::max())||
      (ll_sec<(TmpInt)std::numeric_limits<Wy_Second>::min())) {
     WY_RETURN(Wym_ERANGE);
   }
   WY_RETURN( res.reset(Wy_Second(ll_sec),Wy_Nano(ll_nsec)) );
 }
};

WyRet Wy::_strnum(WyTimeSpec& res, const char** endptr, WyCSeg cseg, const int& radix)
{
  if((radix!=10)&&(radix!=0)) {
    WY_RETURN( Wym_EINVAL );
  }
  return _strnum(res,endptr,cseg); 
};

WyRet Wy::_strnum(WyTimeSpec& res, const char** endptr, WyCSeg cseg, int& radix)
{
  if(radix!=10) {
    if(radix!=0) {
      WY_RETURN( Wym_EINVAL );
    }
    radix=10;
  }
  return _strnum(res,endptr,cseg); 
};
